from apipeline.frames import *
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor

from achatbot.types.frames.data_frames import VisionImageVoiceRawFrame


class VisionImageAudioFrameAggregator(FrameProcessor):
    def __init__(self, pass_audio: bool = False):
        super().__init__()
        self._pass_audio = pass_audio
        self._audio_frame = None

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, AudioRawFrame):
            self._audio_frame = frame
            if self._pass_audio:
                await self.push_frame(frame, direction)
        elif isinstance(frame, VisionImageVoiceRawFrame):
            await self.push_frame(frame, direction)
        elif isinstance(frame, ImageRawFrame):
            if self._audio_frame:
                frame = VisionImageVoiceRawFrame(text=None, audio=self._audio_frame, images=[frame])
                await self.push_frame(frame)
                self._audio_frame = None
        else:
            await self.push_frame(frame, direction)



import os
import logging

import unittest

from src.common.logger import Logger
from src.common.types import LOG_DIR

r"""
python -m unittest test.common.test_logger
"""


class TestLogger(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(logging.TRACE)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_logger_print(self):
        logging.trace("test trace")
        logging.debug("test debug")
        logging.info("test info")
        logging.warning("test warning")
        logging.error("test error")
        logging.critical("test critical")
        logging.exception("test exception")



import logging
import re

from apipeline.frames.data_frames import Frame, ImageRawFrame, TextFrame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor

from achatbot.types.frames.data_frames import LLMMessagesFrame, VisionImageRawFrame


class VisionImageFrameAggregator(FrameProcessor):
    """This aggregator waits for a consecutive TextFrame and an
    ImageFrame. After the ImageFrame arrives it will output a VisionImageFrame.

    >>> async def print_frames(aggregator, frame):
    ...     async for frame in aggregator.process_frame(frame):
    ...         print(frame)

    >>> aggregator = VisionImageFrameAggregator()
    >>> asyncio.run(print_frames(aggregator, TextFrame("What do you see?")))
    >>> asyncio.run(print_frames(aggregator, ImageFrame(image=bytes([]), size=(0, 0))))
    VisionImageFrame, text: What do you see?, image size: 0x0, buffer size: 0 B

    !NOTE: no llm chat history messages, just process raw TextFrame and ImageRawFrame
    """

    def __init__(self, pass_text: bool = False):
        super().__init__()
        self._describe_text = None
        self._pass_text = pass_text

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, TextFrame):
            # maybe use HiTextFrame
            match = re.search(r"\[HI_TEXT\](.*?)\[/HI_TEXT\]", frame.text)
            if match:
                frame.text = match.group(1).strip()
                await self.push_frame(frame, direction)
            else:
                self._describe_text = frame.text
                if self._pass_text:
                    await self.push_frame(frame, direction)
        elif isinstance(frame, VisionImageRawFrame):
            await self.push_frame(frame, direction)
        elif isinstance(frame, ImageRawFrame):
            if self._describe_text:
                frame = VisionImageRawFrame(
                    image=frame.image,
                    size=frame.size,
                    format=frame.format,
                    mode=frame.mode,
                    text=self._describe_text,
                )
                await self.push_frame(frame)
                self._describe_text = None
        else:
            await self.push_frame(frame, direction)



import modal
import os


class ContainerRuntimeConfig:
    images = {
        "default": (
            modal.Image.debian_slim(python_version="3.11")
            .apt_install("git", "git-lfs", "ffmpeg", "sox")
            .pip_install(
                [
                    "achatbot["
                    "fastapi_bot_server,"
                    "livekit,livekit-api,daily,agora,"
                    "silero_vad_analyzer,"
                    "sense_voice_asr,deepgram_asr_processor,"
                    "tts_edge,"
                    "queue"
                    "]~=0.0.8.13",
                    "huggingface_hub[hf_transfer]==0.24.7",
                ],
                extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
            )
            .run_commands(
                f"pip install achatbot[step_voice_processor]==0.0.8.13 --extra-index-url {os.getenv('EXTRA_INDEX_URL', 'https://pypi.org/simple/')}"
            )
            .env(
                {
                    "HF_HUB_ENABLE_HF_TRANSFER": "1",
                    "ACHATBOT_PKG": "1",
                    # asr module engine TAG, default whisper_timestamped_asr
                    "ASR_TAG": "sense_voice_asr",
                    "ASR_LANG": "zn",
                    "ASR_MODEL_NAME_OR_PATH": "/root/.achatbot/models/FunAudioLLM/SenseVoiceSmall",
                    "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
                    "IMAGE_NAME": os.getenv("IMAGE_NAME", "default"),
                    "ACHATBOT_TASK_DONE_TIMEOUT": os.getenv("ACHATBOT_TASK_DONE_TIMEOUT", "1200"),
                    "STEP_TTS_ASSETS_DIR": "/root/.achatbot/assets/",
                }
            )
        ),
    }

    @staticmethod
    def get_img(image_name: str = None):
        image_name = image_name or os.getenv("IMAGE_NAME", "default")
        if image_name not in ContainerRuntimeConfig.images:
            raise Exception(f"image name {image_name} not found")
        print(f"use image:{image_name}")
        return ContainerRuntimeConfig.images[image_name]

    @staticmethod
    def get_app_name(image_name: str = None):
        image_name = image_name or os.getenv("IMAGE_NAME", "default")
        app_name = "fastapi_webrtc_step_voice_bot"
        if image_name != "default":
            app_name = f"fastapi_webrtc_step_voice_{image_name}_bot"
        print(f"app_name:{app_name}")
        return app_name

    @staticmethod
    def get_gpu():
        # T4, L4, A10G, A100, H100
        gpu = os.getenv("IMAGE_GPU", "T4")
        print(f"image_gpu:{gpu}")
        return gpu

    @staticmethod
    def get_allow_concurrent_inputs():
        # T4, L4, A10G, A100, H100
        concurrent_cn = int(os.getenv("IMAGE_CONCURRENT_CN", "1"))
        print(f"image_concurrent_cn:{concurrent_cn}")
        return concurrent_cn


img = ContainerRuntimeConfig.get_img()
with img.imports():
    import logging
    import os

    from achatbot.common.logger import Logger

MODEL_DIR = "/root/.achatbot/models"
model_dir = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assert_dir = modal.Volume.from_name("assets", create_if_missing=True)

# ----------------------- app -------------------------------
app = modal.App("fastapi_webrtc_step_voice_bot")

# volume = modal.Volume.from_name("bot_config", create_if_missing=True)

image = ContainerRuntimeConfig.get_img()
gpu = ContainerRuntimeConfig.get_gpu()
allow_concurrent_inputs = ContainerRuntimeConfig.get_allow_concurrent_inputs()


# 128 MiB of memory and 0.125 CPU cores by default container runtime
@app.cls(
    image=image,
    volumes={MODEL_DIR: model_dir, ASSETS_DIR: assert_dir},
    gpu=gpu,
    secrets=[modal.Secret.from_name("achatbot")],
    cpu=2.0,
    scaledown_window=300,
    timeout=600,
    allow_concurrent_inputs=allow_concurrent_inputs,
)
class Srv:
    # run download_models.py to download models to volume
    # @modal.build()
    # def setup(self):
    #     Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)
    #     # https://huggingface.co/docs/huggingface_hub/guides/download
    #     from huggingface_hub import snapshot_download
    #     from achatbot.common.types import MODELS_DIR

    #     os.makedirs(MODELS_DIR, exist_ok=True)
    #     logging.info(f"start downloading model to dir:{MODELS_DIR}")

    #     # asr model repo
    #     if "sense_voice_asr" in os.getenv("ASR_TAG", "sense_voice_asr"):
    #         local_dir = os.path.join(MODELS_DIR, "FunAudioLLM/SenseVoiceSmall")
    #         snapshot_download(
    #             repo_id="FunAudioLLM/SenseVoiceSmall",
    #             repo_type="model",
    #             allow_patterns="*",
    #             local_dir=local_dir,
    #         )
    #         logging.info(f"sense_voice_asr model to dir:{local_dir} done")

    #     for repo_id in [
    #         "stepfun-ai/Step-Audio-Tokenizer",
    #         "stepfun-ai/Step-Audio-TTS-3B",
    #         "stepfun-ai/Step-Audio-Chat",
    #     ]:
    #         logging.info(f"{repo_id} model to dir:{MODEL_DIR}")
    #         snapshot_download(
    #             repo_id=repo_id,
    #             repo_type="model",
    #             allow_patterns="*",
    #             local_dir=os.path.join(MODEL_DIR, repo_id),
    #         )
    #         logging.info(f"{repo_id} model to dir:{MODEL_DIR} done")

    #     print("download model done")

    @modal.enter()
    def enter(self):
        print("enter done")
        os.makedirs(MODEL_DIR, exist_ok=True)
        os.makedirs(ASSETS_DIR, exist_ok=True)
        # modal.Volume.reload()
        pass

    @modal.asgi_app()
    def app(self):
        from achatbot.cmd.http.server.fastapi_daily_bot_serve import app as fastapi_app

        return fastapi_app



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

"""
Daily REST Helpers

Methods that wrap the Daily API to create rooms, check room URLs, and get meeting tokens.

"""

# TODO: use async aiohttp @weedge
import requests
import time

from urllib.parse import urlparse

from pydantic import Field, BaseModel, ValidationError
from typing import Literal, Optional


class DailyRoomSipParams(BaseModel):
    display_name: str = "sw-sip-dialin"
    video: bool = False
    sip_mode: str = "dial-in"
    num_endpoints: int = 1


class DailyRoomProperties(BaseModel, extra="allow"):
    exp: float = Field(default_factory=lambda: time.time() + 5 * 60)
    enable_chat: bool = False
    enable_emoji_reactions: bool = False
    eject_at_room_exp: bool = True
    enable_prejoin_ui: bool = False  # Important for the bot to be able to join headlessly
    enable_dialout: Optional[bool] = None
    sip: Optional[DailyRoomSipParams] = None
    sip_uri: Optional[dict] = None

    @property
    def sip_endpoint(self) -> str:
        if not self.sip_uri:
            return ""
        else:
            return "sip:%s" % self.sip_uri["endpoint"]


class DailyRoomParams(BaseModel):
    name: Optional[str] = None
    privacy: Literal["private", "public"] = "public"
    properties: DailyRoomProperties = DailyRoomProperties()


class DailyRoomObject(BaseModel):
    id: str
    name: str
    api_created: bool
    privacy: str
    url: str
    created_at: str
    config: DailyRoomProperties


class TokenObject(BaseModel):
    exp: int
    room_name: str
    is_owner: bool
    nbf: Optional[int] = None
    enable_screenshare: Optional[bool] = None
    enable_recording: Optional[str] = None


class DailyRESTHelper:
    def __init__(self, daily_api_key: str, daily_api_url: str = "https://api.daily.co/v1"):
        self.daily_api_key = daily_api_key
        self.daily_api_url = daily_api_url

    @staticmethod
    def get_name_from_url(room_url: str) -> str:
        return urlparse(room_url).path[1:]

    def create_room(self, params: DailyRoomParams) -> DailyRoomObject:
        res = requests.post(
            f"{self.daily_api_url}/rooms",
            headers={"Authorization": f"Bearer {self.daily_api_key}"},
            json={**params.model_dump(exclude_none=True)},
        )

        if res.status_code != 200:
            raise Exception(f"Unable to create room: {res.text}")

        data = res.json()

        try:
            room = DailyRoomObject(**data)
        except ValidationError as e:
            raise Exception(f"Invalid response: {e}")

        return room

    def get_room_from_name(self, room_name: str) -> DailyRoomObject:
        res: requests.Response = requests.get(
            f"{self.daily_api_url}/rooms/{room_name}",
            headers={"Authorization": f"Bearer {self.daily_api_key}"},
        )

        if res.status_code != 200:
            raise Exception(f"Room not found: {room_name}")

        data = res.json()

        try:
            room = DailyRoomObject(**data)
        except ValidationError as e:
            raise Exception(f"Invalid response: {e}")

        return room

    def get_room_from_url(
        self,
        room_url: str,
    ) -> DailyRoomObject:
        room_name = DailyRESTHelper.get_name_from_url(room_url)
        return self.get_room_from_name(room_name)

    def get_token_by_url(
        self, room_url: str, expiry_time: float = 60 * 60, owner: bool = True
    ) -> str:
        if not room_url:
            raise Exception(
                "No Daily room specified. You must specify a Daily room in order a token to be generated."
            )

        room_name = DailyRESTHelper.get_name_from_url(room_url)

        return self.get_token_by_name(room_name, expiry_time=expiry_time, owner=owner)

    def get_token_by_name(
        self, room_name: str, expiry_time: float = 60 * 60, owner: bool = True
    ) -> str:
        expiration: float = time.time() + expiry_time
        res: requests.Response = requests.post(
            f"{self.daily_api_url}/meeting-tokens",
            headers={"Authorization": f"Bearer {self.daily_api_key}"},
            json={"properties": {"room_name": room_name, "is_owner": owner, "exp": expiration}},
        )

        if res.status_code != 200:
            raise Exception(f"Failed to create meeting token: {res.status_code} {res.text}")

        token: str = res.json()["token"]

        return token

    def verify_token(self, token: str) -> TokenObject:
        if not token:
            return False

        res: requests.Response = requests.get(
            f"{self.daily_api_url}/meeting-tokens/{token}",
            headers={"Authorization": f"Bearer {self.daily_api_key}"},
        )

        if res.status_code != 200:
            raise Exception(f"Token not found: {token}")

        data = res.json()

        try:
            print("data--->", data)
            token = TokenObject(**data)
        except ValidationError as e:
            raise Exception(f"Invalid response: {e}")

        return token



# https://nvidia.github.io/TensorRT-LLM/index.html (nice doc)
"""
NeMo -------------
                  |
HuggingFace ------
                  |   convert                             build                    load
Modelopt ---------  ----------> TensorRT-LLM Checkpoint --------> TensorRT Engine ------> TensorRT-LLM ModelRunner
                  |
JAX --------------
                  |
DeepSpeed --------
"""

import os
import modal

app_name = "qwen2.5-0.5B"

app = modal.App("trtllm-generator")

trtllm_image = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.8.0-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install(
        "git", "git-lfs", "openmpi-bin", "libopenmpi-dev", "wget"
    )  # OpenMPI for distributed communication
    .pip_install(
        "tensorrt-llm==0.17.0.post1",
        # "pynvml<12",  # avoid breaking change to pynvml version API for tensorrt_llm
        # "tensorrt==10.8.0.43",
        pre=True,
        extra_index_url="https://pypi.nvidia.com",
    )
    .env({"TORCH_CUDA_ARCH_LIST": "8.0 8.6 8.7 8.9 9.0"})
)

achatbot_trtllm_image = trtllm_image.pip_install(
    "achatbot==0.0.9.post2",
    extra_index_url="https://pypi.org/simple/",
).env(
    {
        "TLLM_LLMAPI_BUILD_CACHE": "1",
    }
)

MAX_BATCH_SIZE = 1024  # better throughput at larger batch sizes, limited by GPU RAM
MODEL_ID = "Qwen/Qwen2.5-0.5B"

HF_MODEL_DIR = "/root/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
TRT_MODEL_DIR = "/root/trt_models"
trt_model_vol = modal.Volume.from_name("triton_trtllm_models", create_if_missing=True)

TRT_MODEL_CACHE_DIR = "/tmp/.cache/tensorrt_llm/llmapi/"
trt_model_cache_vol = modal.Volume.from_name("triton_trtllm_cache_models", create_if_missing=True)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=trtllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_CACHE_DIR: trt_model_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
def run_sync():
    """
    https://nvidia.github.io/TensorRT-LLM/_modules/tensorrt_llm/llmapi/llm.html#LLM.generate
    """
    from tensorrt_llm import LLM, SamplingParams
    from tensorrt_llm.bindings.executor import KvCacheConfig

    prompts = [
        "Hello, my name is",
        "The president of the United States is",
        "The capital of France is",
        "The future of AI is",
    ]
    # https://nvidia.github.io/TensorRT-LLM/llm-api/reference.html#tensorrt_llm.llmapi.SamplingParams
    sampling_params = SamplingParams(temperature=0.8, top_p=0.95)

    # https://github.com/NVIDIA/TensorRT-LLM/blob/main/tensorrt_llm/llmapi/llm_args.py#L520
    kv_cache_config=KvCacheConfig(free_gpu_memory_fraction=0.5)
    # load hf model, convert to tensorrt, build tensorrt engine, load tensorrt engine
    llm = LLM(model=MODEL_ID, kv_cache_config=kv_cache_config)

    outputs = llm.generate(prompts, sampling_params, use_tqdm=False)

    # Print the outputs.
    for output in outputs:
        print(output)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=trtllm_image.env({}),
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_CACHE_DIR: trt_model_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_async_stream():
    """
    https://nvidia.github.io/TensorRT-LLM/_modules/tensorrt_llm/llmapi/llm.html#LLM.generate_async
    """
    from tensorrt_llm import LLM, SamplingParams

    prompts = [
        "Hello, my name is",
        # "The president of the United States is",
        # "The capital of France is",
        # "The future of AI is",
    ]
    # https://nvidia.github.io/TensorRT-LLM/llm-api/reference.html#tensorrt_llm.llmapi.SamplingParams
    sampling_params = SamplingParams(temperature=0.8, top_p=0.95, max_tokens=2, detokenize=False)

    # load hf model, convert to tensorrt, build tensorrt engine, load tensorrt engine
    llm = LLM(model=MODEL_ID)

    for i, prompt in enumerate(prompts):
        generator = llm.generate_async(prompt, sampling_params, streaming=True)
        async for output in generator:
            print(output)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=trtllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_CACHE_DIR: trt_model_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_async_batch_stream():
    """
    https://nvidia.github.io/TensorRT-LLM/_modules/tensorrt_llm/llmapi/llm.html#LLM.generate_async
    """
    import asyncio
    import uuid

    from tensorrt_llm import LLM, SamplingParams

    # Prompts to generate
    prompts = [
        "Hello, my name is",
        "The president of the United States is",
        "The capital of France is",
        "The future of AI is",
    ]

    # Sampling parameters
    sampling_params = SamplingParams(
        temperature=0.8, top_p=0.95, max_tokens=2, detokenize=True, return_perf_metrics=True
    )

    # Load HF model, convert to TensorRT, build TensorRT engine, load TensorRT engine
    llm = LLM(model=MODEL_ID)

    lock = asyncio.Lock()

    async def run_async_stream(llm, prompt, sampling_params, request_id=str(uuid.uuid4().hex)):
        generator = llm.generate_async(prompt, sampling_params, streaming=True)
        async for item in generator:
            print(f"[{request_id}] tokenId: {item.outputs[0].token_ids[-1]} {item} ")
            async with lock:
                print(f"[{request_id}] {item}")
                print(item.outputs[0].token_ids[-1])
                # u can send this response to a request queue/channle

    tasks = [
        run_async_stream(llm, prompt, sampling_params, request_id=str(uuid.uuid4().hex))
        for prompt in prompts
    ]
    await asyncio.gather(*tasks)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=trtllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_DIR: trt_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
def runner_stream():
    """
    https://github.com/NVIDIA/TensorRT-LLM/blob/main/examples/run.py#L548
    """
    import time

    import tensorrt_llm
    from tensorrt_llm.runtime import ModelRunner, SamplingConfig
    from transformers import AutoTokenizer

    init_start = time.monotonic_ns()

    tokenizer = AutoTokenizer.from_pretrained(os.path.join(HF_MODEL_DIR, MODEL_ID))
    # tokenizer.pad_token = tokenizer.eos_token
    # tokenizer.add_special_tokens({"pad_token": tokenizer.eos_token})
    # and then we add it from the left, to minimize impact on the output
    # tokenizer.padding_side = "left"
    pad_id = tokenizer.pad_token_id
    end_id = tokenizer.eos_token_id
    print(f"pad {tokenizer.pad_token} {pad_id}")
    print(f"eos {tokenizer.eos_token} {end_id}")

    prompt_cases = [
        {"prompt": "hello, my name is", "kwargs": {"max_new_tokens": 10, "stop_words_list": None}},
        {
            "prompt": "hello, my name is",
            "kwargs": {
                "max_new_tokens": 10,
                "stop_words_list": [[[end_id]]],  # one prompt batch stop ids
            },
        },  # prefill cache token test (default no cache)
        {
            "prompt": "hello, what your name?",
            "kwargs": {
                "max_new_tokens": 10,
            },
        },
    ]

    # load tensorrt engine
    # https://nvidia.github.io/TensorRT-LLM/python-api/tensorrt_llm.runtime.html#tensorrt_llm.runtime.ModelRunner.from_dir
    engine = ModelRunner.from_dir(
        engine_dir=os.path.join(TRT_MODEL_DIR, "qwen2.5-0.5B", "trt_engines_bfloat16"),
        rank=tensorrt_llm.mpi_rank(),  # this will need to be adjusted to use multiple GPUs
        max_output_len=100,
        debug_mode=True,
    )

    init_duration_s = (time.monotonic_ns() - init_start) / 1e9
    print(f"ModelRunner init duration: {init_duration_s:.2f} s")

    for i, case in enumerate(prompt_cases):
        # https://nvidia.github.io/TensorRT-LLM/_modules/tensorrt_llm/runtime/generation.html#SamplingConfig
        sampling_config = SamplingConfig(
            end_id=end_id,
            pad_id=pad_id,
            temperature=0.8,
            repetition_penalty=1.1,
            max_new_tokens=50,
            top_k=20,
            top_p=0.95,
            # min_p version > 0.17.0
            # min_p=0.0,
            stop_words_list=None,
        )
        sampling_config.update(**case["kwargs"])
        print("sampling_config:", sampling_config)
        tokens = tokenizer([case["prompt"]], return_tensors="pt", padding=False, truncation=False)
        input_ids = tokens["input_ids"]
        print("input_ids:", input_ids)
        # https://nvidia.github.io/TensorRT-LLM/python-api/tensorrt_llm.runtime.html#tensorrt_llm.runtime.ModelRunner.generate
        generator = engine.generate(
            input_ids,
            streaming=True,
            sampling_config=sampling_config,
        )
        first = True
        start_time = time.perf_counter()
        for output in generator:
            if first:
                ttft = time.perf_counter() - start_time
                print(f"generate TTFT time: {ttft} s")
                first = False
            print(output)
            output = output[:, 0][0]
            # mask = output != pad_id
            # output = output[mask]
            outputs_text = tokenizer.decode(output, skip_special_tokens=True)
            print(output, outputs_text)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=trtllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_DIR: trt_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
def runner_batch_stream():
    """
    https://github.com/NVIDIA/TensorRT-LLM/blob/main/examples/run.py#L548
    """
    import time
    import random

    import tensorrt_llm
    from tensorrt_llm.runtime import ModelRunner, SamplingConfig
    from transformers import AutoTokenizer

    # from trtllm-build --max_batch_size 16
    # better throughput at larger batch sizes, but limited by GPU RAM
    MAX_BATCH_SIZE = 16

    init_start = time.monotonic_ns()

    tokenizer = AutoTokenizer.from_pretrained(os.path.join(HF_MODEL_DIR, MODEL_ID))
    # tokenizer.pad_token = tokenizer.eos_token
    # tokenizer.add_special_tokens({"pad_token": tokenizer.eos_token})
    # and then we add it from the left, to minimize impact on the output
    # tokenizer.padding_side = "left"
    pad_id = tokenizer.pad_token_id
    end_id = tokenizer.eos_token_id
    print(f"pad {tokenizer.pad_token} {pad_id}")
    print(f"pad {tokenizer.eos_token} {end_id}")

    questions = [
        # Generic assistant questions
        "What are you?",
        "What can you do?",
        # Coding
        "Implement a Python function to compute the Fibonacci numbers.",
        "Write a Rust function that performs binary exponentiation.",
        "How do I allocate memory in C?",
        "What are the differences between Javascript and Python?",
        "How do I find invalid indices in Postgres?",
        "How can you implement a LRU (Least Recently Used) cache in Python?",
        "What approach would you use to detect and prevent race conditions in a multithreaded application?",
        "Can you explain how a decision tree algorithm works in machine learning?",
        "How would you design a simple key-value store database from scratch?",
        "How do you handle deadlock situations in concurrent programming?",
        "What is the logic behind the A* search algorithm, and where is it used?",
        "How can you design an efficient autocomplete system?",
        "What approach would you take to design a secure session management system in a web application?",
        "How would you handle collision in a hash table?",
        "How can you implement a load balancer for a distributed system?",
        "Implement a Python class for a doubly linked list.",
        "Write a Haskell function that generates prime numbers using the Sieve of Eratosthenes.",
        "Develop a simple HTTP server in Rust.",
        # Literate and creative writing
        "What is the fable involving a fox and grapes?",
        "Who does Harry turn into a balloon?",
        "Write a story in the style of James Joyce about a trip to the Australian outback in 2083 to see robots in the beautiful desert.",
        "Write a tale about a time-traveling historian who's determined to witness the most significant events in human history.",
        "Describe a day in the life of a secret agent who's also a full-time parent.",
        "Create a story about a detective who can communicate with animals.",
        "What is the most unusual thing about living in a city floating in the clouds?",
        "In a world where dreams are shared, what happens when a nightmare invades a peaceful dream?",
        "Describe the adventure of a lifetime for a group of friends who found a map leading to a parallel universe.",
        "Tell a story about a musician who discovers that their music has magical powers.",
        "In a world where people age backwards, describe the life of a 5-year-old man.",
        "Create a tale about a painter whose artwork comes to life every night.",
        "What happens when a poet's verses start to predict future events?",
        "Imagine a world where books can talk. How does a librarian handle them?",
        "Tell a story about an astronaut who discovered a planet populated by plants.",
        "Describe the journey of a letter traveling through the most sophisticated postal service ever.",
        "Write a tale about a chef whose food can evoke memories from the eater's past.",
        "Write a poem in the style of Walt Whitman about the modern digital world.",
        "Create a short story about a society where people can only speak in metaphors.",
        "What are the main themes in Dostoevsky's 'Crime and Punishment'?",
        # History and Philosophy
        "What were the major contributing factors to the fall of the Roman Empire?",
        "How did the invention of the printing press revolutionize European society?",
        "What are the effects of quantitative easing?",
        "How did the Greek philosophers influence economic thought in the ancient world?",
        "What were the economic and philosophical factors that led to the fall of the Soviet Union?",
        "How did decolonization in the 20th century change the geopolitical map?",
        "What was the influence of the Khmer Empire on Southeast Asia's history and culture?",
        "What led to the rise and fall of the Mongol Empire?",
        "Discuss the effects of the Industrial Revolution on urban development in 19th century Europe.",
        "How did the Treaty of Versailles contribute to the outbreak of World War II?",
        "What led to the rise and fall of the Mongol Empire?",
        "Discuss the effects of the Industrial Revolution on urban development in 19th century Europe.",
        "How did the Treaty of Versailles contribute to the outbreak of World War II?",
        "Explain the concept of 'tabula rasa' in John Locke's philosophy.",
        "What does Nietzsche mean by 'ressentiment'?",
        "Compare and contrast the early and late works of Ludwig Wittgenstein. Which do you prefer?",
        "How does the trolley problem explore the ethics of decision-making in critical situations?",
        # Thoughtfulness
        "Describe the city of the future, considering advances in technology, environmental changes, and societal shifts.",
        "In a dystopian future where water is the most valuable commodity, how would society function?",
        "If a scientist discovers immortality, how could this impact society, economy, and the environment?",
        "What could be the potential implications of contact with an advanced alien civilization?",
        "Describe how you would mediate a conflict between two roommates about doing the dishes using techniques of non-violent communication.",
        "If you could design a school curriculum for the future, what subjects would you include to prepare students for the next 50 years?",
        "How would society change if teleportation was invented and widely accessible?",
        "Consider a future where artificial intelligence governs countries. What are the potential benefits and pitfalls?",
        # Math
        "What is the product of 9 and 8?",
        "If a train travels 120 kilometers in 2 hours, what is its average speed?",
        "Think through this step by step. If the sequence a_n is defined by a_1 = 3, a_2 = 5, and a_n = a_(n-1) + a_(n-2) for n > 2, find a_6.",
        "Think through this step by step. Calculate the sum of an arithmetic series with first term 3, last term 35, and total terms 11.",
        "Think through this step by step. What is the area of a triangle with vertices at the points (1,2), (3,-4), and (-2,5)?",
        "Think through this step by step. Solve the following system of linear equations: 3x + 2y = 14, 5x - y = 15.",
        # Facts
        "Who was Emperor Norton I, and what was his significance in San Francisco's history?",
        "What is the Voynich manuscript, and why has it perplexed scholars for centuries?",
        "What was Project A119 and what were its objectives?",
        "What is the 'Dyatlov Pass incident' and why does it remain a mystery?",
        "What is the 'Emu War' that took place in Australia in the 1930s?",
        "What is the 'Phantom Time Hypothesis' proposed by Heribert Illig?",
        "Who was the 'Green Children of Woolpit' as per 12th-century English legend?",
        "What are 'zombie stars' in the context of astronomy?",
        "Who were the 'Dog-Headed Saint' and the 'Lion-Faced Saint' in medieval Christian traditions?",
        "What is the story of the 'Globsters', unidentified organic masses washed up on the shores?",
        "Which countries in the European Union use currencies other than the Euro, and what are those currencies?",
        # Multilingual
        "战国时期最重要的人物是谁?",
        "Tuende hatua kwa hatua. Hesabu jumla ya mfululizo wa kihesabu wenye neno la kwanza 2, neno la mwisho 42, na jumla ya maneno 21.",
        "Kannst du die wichtigsten Eigenschaften und Funktionen des NMDA-Rezeptors beschreiben?",
        "¿Cuáles son los principales impactos ambientales de la deforestación en la Amazonía?",
        "Décris la structure et le rôle de la mitochondrie dans une cellule.",
        "Какие были социальные последствия Перестройки в Советском Союзе?",
        # Economics and Business
        "What are the principles of behavioral economics and how do they influence consumer choices?",
        "Discuss the impact of blockchain technology on traditional banking systems.",
        "What are the long-term effects of trade wars on global economic stability?",
        "What is the law of supply and demand?",
        "Explain the concept of inflation and its typical causes.",
        "What is a trade deficit, and why does it matter?",
        "How do interest rates affect consumer spending and saving?",
        "What is GDP and why is it important for measuring economic health?",
        "What is the difference between revenue and profit?",
        "Describe the role of a business plan in startup success.",
        "How does market segmentation benefit a company?",
        "Explain the concept of brand equity.",
        "What are the advantages of franchising a business?",
        "What are Michael Porter's five forces and how do they impact strategy for tech startups?",
        # Science and Technology
        "Discuss the potential impacts of quantum computing on data security.",
        "How could CRISPR technology change the future of medical treatments?",
        "Explain the significance of graphene in the development of future electronics.",
        "How do renewable energy sources compare to fossil fuels in terms of environmental impact?",
        "What are the most promising technologies for carbon capture and storage?",
        "Explain why the sky is blue.",
        "What is the principle behind the operation of a microwave oven?",
        "How does Newton's third law apply to rocket propulsion?",
        "What causes iron to rust?",
        "Describe the process of photosynthesis in simple terms.",
        "What is the role of a catalyst in a chemical reaction?",
        "What is the basic structure of a DNA molecule?",
        "How do vaccines work to protect the body from disease?",
        "Explain the significance of mitosis in cellular reproduction.",
        "What are tectonic plates and how do they affect earthquakes?",
        "How does the greenhouse effect contribute to global warming?",
        "Describe the water cycle and its importance to Earth's climate.",
        "What causes the phases of the Moon?",
        "How do black holes form?",
        "Explain the significance of the Big Bang theory.",
        "What is the function of the CPU in a computer system?",
        "Explain the difference between RAM and ROM.",
        "How does a solid-state drive (SSD) differ from a hard disk drive (HDD)?",
        "What role does the motherboard play in a computer system?",
        "Describe the purpose and function of a GPU.",
        "What is TensorRT? What role does it play in neural network inference?",
    ]

    prefixes = [
        "Hi! ",
        "Hello! ",
        "Hi. ",
        "Hello. ",
        "Hi: ",
        "Hello: ",
        "Greetings. ",
    ]
    # prepending any string that causes a tokenization change is enough to invalidate KV cache
    for ii, prefix in enumerate(prefixes):
        questions += [prefix + question for question in questions[:128]]

    prompt_cases = [
        {
            # "prompts": ["hi, my name is", "hello, my name is"], # test kv cache
            # "prompts": questions, # test batch size
            "prompts": random.sample(questions, MAX_BATCH_SIZE),
            "kwargs": {"max_new_tokens": 10, "stop_words_list": None},
        },
    ]

    # load tensorrt engine
    # https://nvidia.github.io/TensorRT-LLM/python-api/tensorrt_llm.runtime.html#tensorrt_llm.runtime.ModelRunner.from_dir
    engine = ModelRunner.from_dir(
        engine_dir=os.path.join(TRT_MODEL_DIR, "qwen2.5-0.5B", "trt_engines_bfloat16"),
        rank=tensorrt_llm.mpi_rank(),  # this will need to be adjusted to use multiple GPUs
        max_output_len=100,
        debug_mode=False,
    )

    init_duration_s = (time.monotonic_ns() - init_start) / 1e9
    print(f"ModelRunner init duration: {init_duration_s:.2f} s")

    for i, case in enumerate(prompt_cases):
        # https://nvidia.github.io/TensorRT-LLM/_modules/tensorrt_llm/runtime/generation.html#SamplingConfig
        sampling_config = SamplingConfig(
            end_id=end_id,
            pad_id=pad_id,
            temperature=0.8,
            repetition_penalty=1.1,
            max_new_tokens=50,
            top_k=20,
            top_p=0.95,
            # min_p version > 0.17.0
            # min_p=0.0,
            stop_words_list=None,
        )
        sampling_config.update(**case["kwargs"])
        print("sampling_config:", sampling_config)
        tokens = tokenizer(case["prompts"], return_tensors="pt", padding=True, truncation=False)
        input_ids = tokens["input_ids"]
        print("input_ids:", input_ids.shape)

        num_prompts = input_ids.shape[0]
        if num_prompts > MAX_BATCH_SIZE:
            raise ValueError(f"Batch size {num_prompts} exceeds maximum of {MAX_BATCH_SIZE}")

        # https://nvidia.github.io/TensorRT-LLM/python-api/tensorrt_llm.runtime.html#tensorrt_llm.runtime.ModelRunner.generate
        generator = engine.generate(
            input_ids,
            streaming=True,
            sampling_config=sampling_config,
        )
        first = True
        start_time = time.perf_counter()
        for output in generator:
            if first:
                ttft = time.perf_counter() - start_time
                print(f"generate batch_size {output.shape[0]} TTFT time: {ttft} s")
                first = False
            print("raw output", output.shape)
            output = output[:, 0]
            print("output", output.shape)
            # mask = output != pad_id
            # output = output[mask]
            outputs_text = tokenizer.batch_decode(output, skip_special_tokens=True)
            assert len(outputs_text) == output.shape[0]
            print(output.shape, outputs_text)
        batch_stream_generate_cost = time.perf_counter() - start_time
        print(
            f"stream generate batch_size {output.shape[0]} cost time: {batch_stream_generate_cost} s"
        )


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=achatbot_trtllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_CACHE_DIR: trt_model_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_achatbot_generator():
    import uuid
    import os
    import asyncio
    import time

    from achatbot.core.llm.tensorrt_llm.generator import (
        TrtLLMGenerator,
        TensorRTLLMEngineArgs,
        LMGenerateArgs,
        LlmArgs,
    )
    from achatbot.common.types import SessionCtx
    from achatbot.common.session import Session
    from transformers import AutoTokenizer, GenerationConfig

    model = os.path.join(HF_MODEL_DIR, MODEL_ID)
    generator = TrtLLMGenerator(
        **TensorRTLLMEngineArgs(serv_args=LlmArgs(model=model).to_dict()).__dict__,
    )
    tokenizer = AutoTokenizer.from_pretrained(model)
    generation_config = {}
    if os.path.exists(os.path.join(model, "generation_config.json")):
        generation_config = GenerationConfig.from_pretrained(
            model, "generation_config.json"
        ).to_dict()
    # async def run():
    prompt_cases = [
        {"prompt": "hello, my name is", "kwargs": {"max_new_tokens": 30, "stop_ids": []}},
        {
            "prompt": "hello, my name is",
            "kwargs": {"max_new_tokens": 30, "stop_ids": [13]},
        },  # prefill cache token test (default no cache)
        {"prompt": "hello, what your name?", "kwargs": {"max_new_tokens": 30, "stop_ids": [13]}},
    ]

    # test the same session
    # session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
    for case in prompt_cases:
        session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
        tokens = tokenizer(case["prompt"])
        session.ctx.state["token_ids"] = tokens["input_ids"]
        gen_kwargs = {**generation_config, **case["kwargs"], **tokens}
        print("gen_kwargs:", gen_kwargs)
        first = True
        start_time = time.perf_counter()
        gen_texts = ""
        async for token_id in generator.generate(session, **gen_kwargs):
            if first:
                ttft = time.perf_counter() - start_time
                print(f"generate TTFT time: {ttft} s")
                first = False
            gen_text = tokenizer.decode(token_id)
            gen_texts += gen_text
            print(session.ctx.client_id, token_id, gen_text)
        print(session.ctx.client_id, gen_texts)
    # asyncio.run(run())


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    retries=0,
    image=achatbot_trtllm_image,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        TRT_MODEL_DIR: trt_model_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_achatbot_runner_generator():
    import uuid
    import os
    import asyncio
    import time

    from achatbot.core.llm.tensorrt_llm.generator import (
        TrtLLMRunnerGenerator,
        TensorRTLLMRunnerArgs,
        TensorRTLLMRunnerEngineArgs,
        LMGenerateArgs,
    )
    from achatbot.common.types import SessionCtx
    from achatbot.common.session import Session
    from transformers import AutoTokenizer, GenerationConfig

    model = os.path.join(HF_MODEL_DIR, MODEL_ID)
    engine_dir = os.path.join(TRT_MODEL_DIR, "qwen2.5-0.5B", "trt_engines_bfloat16")
    generator = TrtLLMRunnerGenerator(
        **TensorRTLLMRunnerEngineArgs(
            serv_args=TensorRTLLMRunnerArgs(engine_dir=engine_dir).__dict__
        ).__dict__,
    )
    tokenizer = AutoTokenizer.from_pretrained(model)
    generation_config = {}
    if os.path.exists(os.path.join(model, "generation_config.json")):
        generation_config = GenerationConfig.from_pretrained(
            model, "generation_config.json"
        ).to_dict()
    # async def run():
    prompt_cases = [
        {"prompt": "hello, my name is", "kwargs": {"max_new_tokens": 30, "stop_ids": []}},
        {
            "prompt": "hello, my name is",
            "kwargs": {"max_new_tokens": 30, "stop_ids": [13]},
        },  # prefill cache token test (default no cache)
        {"prompt": "hello, what your name?", "kwargs": {"max_new_tokens": 30, "stop_ids": [13]}},
    ]

    # test the same session
    # session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
    for case in prompt_cases:
        session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
        tokens = tokenizer(case["prompt"])
        session.ctx.state["token_ids"] = tokens["input_ids"]
        # gen_kwargs = {**generation_config, **case["kwargs"], **tokens}
        gen_kwargs = {**case["kwargs"], **tokens}
        print("gen_kwargs:", gen_kwargs)
        first = True
        start_time = time.perf_counter()
        gen_texts = ""
        async for token_id in generator.generate(session, **gen_kwargs):
            if first:
                ttft = time.perf_counter() - start_time
                print(f"generate TTFT time: {ttft} s")
                first = False
            gen_text = tokenizer.decode(token_id)
            gen_texts += gen_text
            print(session.ctx.client_id, token_id, gen_text)
        print(session.ctx.client_id, gen_texts)
    # asyncio.run(run())


"""
# llmapi (LLM load hf model, convert to tensorrt, build tensorrt engine, load tensorrt engine)
modal run src/llm/trtllm/examples/generator.py::run_sync (no stream | batch prompts processing | throughput)
modal run src/llm/trtllm/examples/generator.py::run_async_stream (stream | single prompt async processing | latency)
modal run src/llm/trtllm/examples/generator.py::run_async_batch_stream (stream | batch prompts async processing | latency+throughput) (multiple prompts/request or one prompt/request)

# runner (load tensorrt engine to run generate)
modal run src/llm/trtllm/examples/generator.py::runner_stream
modal run src/llm/trtllm/examples/generator.py::runner_batch_stream

# achatbot
modal run src/llm/trtllm/examples/generator.py::run_achatbot_generator
modal run src/llm/trtllm/examples/generator.py::run_achatbot_runner_generator

"""



# fuck magic number

ROOM_EXPIRE_TIME = 30 * 60  # 30 minutes
ROOM_TOKEN_EXPIRE_TIME = 30 * 60  # 30 minutes
RANDOM_ROOM_EXPIRE_TIME = 5 * 60  # 5 minutes
MAX_RANDOM_ROOM_EXPIRE_TIME = 30 * 60  # 30 minutes
RANDOM_ROOM_TOKEN_EXPIRE_TIME = 5 * 60  # 5 minutes
DAILYLANGCHAINRAGBOT_EXPIRE_TIME = 25 * 60


# VideoFrame type
# https://api-ref.agora.io/en/mediaplayer-kit/windows/1.x/structagora_1_1media_1_1base_1_1_video_frame.html (0-16)
# ExternalVideoFrame format
# https://api-ref.agora.io/en/voice-sdk/unity/4.x/API/enum_videopixelformat.html (0,1,4,16,17)
# video pixel format in Agora SDK
AGORA_VIDEO_PIXEL_UNKNOWN = 0  # The format is known.
AGORA_VIDEO_PIXEL_I420 = 1  # The format is I420. YUV 420
AGORA_VIDEO_PIXEL_BGRA = 2  # The format is BGRA.
AGORA_VIDEO_PIXEL_NV21 = 3  # The format is NV21.
AGORA_VIDEO_PIXEL_RGBA = 4  # The format is RGBA.
AGORA_VIDEO_PIXEL_NV12 = 8  # The format is NV12.
AGORA_VIDEO_TEXTURE_2D = 10  # The format is GL_TEXTURE_2D.
AGORA_VIDEO_TEXTURE_OES = 11  # The format is GL_TEXTURE_OES.
AGORA_VIDEO_PIXEL_I422 = 16  # The format is I422. YUV 422
AGORA_VIDEO_TEXTURE_ID3D11TEXTURE2D = 17

# ExternalVideoFrame type (buffer type)
# https://api-ref.agora.io/en/voice-sdk/unity/4.x/API/enum_videobuffertype.html
# The video buffer in the format
AGORA_VIDEO_BUFFER_RAW_DATA = 1  # raw data.
AGORA_VIDEO_BUFFER_ARRAY = 2  # raw data array.
AGORA_VIDEO_BUFFER_TEXTURE = 3  # Texture.



r"""
pyannote
- https://www.pyannote.ai/
- https://github.com/pyannote/pyannote-audio
hf-repo-ckpt:
- https://huggingface.co/pyannote/segmentation
- https://huggingface.co/pyannote/segmentation-3.0
- https://huggingface.co/pyannote/speaker-diarization-3.0
- https://huggingface.co/pyannote/speaker-diarization-3.1

paper:
- [pyannote.audio: neural building blocks for speaker diarization](https://arxiv.org/abs/1911.01255)
- [End-to-end speaker segmentation for overlap-aware resegmentation](https://arxiv.org/abs/2104.04045)
- [Powerset multi-class cross entropy loss for neural speaker diarization](https://arxiv.org/abs/2310.13025)
- [PYANNOTE.AUDIO 2.1 SPEAKER DIARIZATION PIPELINE: PRINCIPLE, BENCHMARK, AND RECIPE](https://huggingface.co/paris-iea/speaker-diarization/resolve/main/technical_report_2.1.pdf)
"""

import os


def init_model(
    hf_auth_token,
    path_or_hf_repo="pyannote/segmentation-3.0",
):
    from pyannote.audio import Model

    auth_token = os.environ.get("HF_TOKEN") if os.environ.get("HF_TOKEN") else hf_auth_token

    # 1. visit hf.co/pyannote/segmentation-3.0 and accept user conditions
    # 2. visit hf.co/settings/tokens to create an access token
    # 3. instantiate pretrained model
    model = Model.from_pretrained(path_or_hf_repo, use_auth_token=auth_token)

    return model


def pyannote_vad_pipeline(
    audio_path,
    hf_auth_token,
    path_or_hf_repo="pyannote/segmentation-3.0",
    model_type="segmentation-3.0",
):
    r"""
    voice activity detection (语音活动识别)
    """
    from pyannote.audio.pipelines import VoiceActivityDetection

    pipeline = VoiceActivityDetection(segmentation=init_model(hf_auth_token, path_or_hf_repo))
    HYPER_PARAMETERS = {
        # remove speech regions shorter than that many seconds.
        "min_duration_on": 0.0,
        # fill non-speech regions shorter than that many seconds.
        "min_duration_off": 0.0,
    }
    # if use pyannote/segmentation open onset/offset activation thresholds
    if model_type == "segmentation":
        HYPER_PARAMETERS["onset"] = 0.5
        HYPER_PARAMETERS["offset"] = 0.5
    pipeline.instantiate(HYPER_PARAMETERS)
    vad_res = pipeline(audio_path)
    print(type(vad_res), vad_res)
    # `vad_res` is a pyannote.core.Annotation instance containing speech regions
    for segment in vad_res.itersegments():
        print(type(segment), segment, segment.start, segment.end)


def pyannote_osd_pipeline(
    audio_path,
    hf_auth_token,
    path_or_hf_repo="pyannote/segmentation-3.0",
    model_type="segmentation-3.0",
):
    r"""
    Overlapped speech detection (重叠语音检测)
    """
    from pyannote.audio.pipelines import OverlappedSpeechDetection

    pipeline = OverlappedSpeechDetection(segmentation=init_model(hf_auth_token, path_or_hf_repo))
    HYPER_PARAMETERS = {
        # remove speech regions shorter than that many seconds.
        "min_duration_on": 0.0,
        # fill non-speech regions shorter than that many seconds.
        "min_duration_off": 0.0,
    }
    # if use pyannote/segmentation open onset/offset activation thresholds
    if model_type == "segmentation":
        HYPER_PARAMETERS["onset"] = 0.5
        HYPER_PARAMETERS["offset"] = 0.5
    pipeline.instantiate(HYPER_PARAMETERS)
    vad_res = pipeline(audio_path)
    print(type(vad_res), vad_res)
    # `vad_res` is a pyannote.core.Annotation instance containing speech regions
    for segment in vad_res.itersegments():
        print(type(segment), segment, segment.start, segment.end)


def pyannote_diarization_pipeline(
    audio_path,
    hf_auth_token,
    path_or_hf_repo="pyannote/speaker-diarization-3.1",
    diarization_path="./records/diarization_audio.rttm",
):
    r"""
    Speaker diarization (说话人分割或说话人辨识)
    """
    from pyannote.audio import Pipeline

    auth_token = os.environ.get("HF_TOKEN") if os.environ.get("HF_TOKEN") else hf_auth_token
    # instantiate the pipeline
    pipeline = Pipeline.from_pretrained(path_or_hf_repo, use_auth_token=auth_token)

    # run the pipeline on an audio file
    diarization = pipeline(audio_path)
    print(type(diarization), diarization)

    # Pre-loading audio files in memory may result in faster processing:
    # import torchaudio
    # waveform, sample_rate = torchaudio.load("audio.wav")
    # diarization = pipeline({"waveform": waveform, "sample_rate": sample_rate})

    # Monitoring progress Hooks are available to monitor the progress of the pipeline:
    # from pyannote.audio.pipelines.utils.hook import ProgressHook
    # with ProgressHook() as hook:
    #    diarization = pipeline("audio.wav", hook=hook)

    # Controlling the number of speakers
    # diarization = pipeline("audio.wav", num_speakers=2)
    # diarization = pipeline("audio.wav", min_speakers=2, max_speakers=5)

    for turn, _, speaker in diarization.itertracks(yield_label=True):
        print(type(turn), turn, f"start: {turn.start:.3f}", f"end: {turn.end:.3f}")
        print(type(speaker), speaker)

    # dump the diarization output to disk using RTTM format
    with open(diarization_path, "w") as rttm:
        diarization.write_rttm(rttm)


if __name__ == "__main__":
    r"""
    python demo/vad_pyannote.py -m ./models/pyannote/segmentation/pytorch_model.bin -mt segmentation
    python demo/vad_pyannote.py -dt osd
    python demo/vad_pyannote.py -m pyannote/speaker-diarization-3.1 -dt diarization -dp ./records/diarization_audio.rttm
    """
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--audio_path", "-a", type=str, default="./records/tmp.wav", help="audio path"
    )
    parser.add_argument("--auth_token", "-t", type=str, default="", help="hf auth token")
    parser.add_argument(
        "--path_or_hf_repo",
        "-m",
        type=str,
        default="./models/pyannote/segmentation-3.0/pytorch_model.bin",
        help="model ckpt file path or hf repo",
    )
    parser.add_argument(
        "--model_type",
        "-mt",
        type=str,
        default="segmentation-3.0",
        choices=["segmentation-3.0", "segmentation", "diarization"],
        help="choice segmentation or segmentation-3.0 or diarization",
    )
    parser.add_argument(
        "--detect_type",
        "-dt",
        type=str,
        default="vad",
        choices=["vad", "osd", "diarization"],
        help="choice vad, osd, diarization",
    )
    parser.add_argument(
        "--diarization_path",
        "-dp",
        type=str,
        default="./records/diarization_audio.rttm",
        help="diarization rttm file path",
    )
    args = parser.parse_args()

    if args.detect_type == "diarization":
        pyannote_diarization_pipeline(
            args.audio_path, args.auth_token, args.path_or_hf_repo, args.diarization_path
        )
    elif args.detect_type == "osd":
        pyannote_osd_pipeline(
            args.audio_path, args.auth_token, args.path_or_hf_repo, args.model_type
        )
    else:
        pyannote_vad_pipeline(
            args.audio_path, args.auth_token, args.path_or_hf_repo, args.model_type
        )



# Copyright 2019 Hitachi, Ltd. (author: Yusuke Fujita)
# Licensed under the MIT license.
#
# This library provides utilities for kaldi-style data directory.


from __future__ import print_function
import os
import sys
import numpy as np
import subprocess
import librosa as sf
import io
from functools import lru_cache


def load_segments(segments_file):
    """load segments file as array"""
    if not os.path.exists(segments_file):
        return None
    return np.loadtxt(
        segments_file,
        dtype=[("utt", "object"), ("rec", "object"), ("st", "f"), ("et", "f")],
        ndmin=1,
    )


def load_segments_hash(segments_file):
    ret = {}
    if not os.path.exists(segments_file):
        return None
    for line in open(segments_file):
        utt, rec, st, et = line.strip().split()
        ret[utt] = (rec, float(st), float(et))
    return ret


def load_segments_rechash(segments_file):
    ret = {}
    if not os.path.exists(segments_file):
        return None
    for line in open(segments_file):
        utt, rec, st, et = line.strip().split()
        if rec not in ret:
            ret[rec] = []
        ret[rec].append({"utt": utt, "st": float(st), "et": float(et)})
    return ret


def load_wav_scp(wav_scp_file):
    """return dictionary { rec: wav_rxfilename }"""
    lines = [line.strip().split(None, 1) for line in open(wav_scp_file)]
    return {x[0]: x[1] for x in lines}


@lru_cache(maxsize=1)
def load_wav(wav_rxfilename, start=0, end=None):
    """This function reads audio file and return data in numpy.float32 array.
    "lru_cache" holds recently loaded audio so that can be called
    many times on the same audio file.
    OPTIMIZE: controls lru_cache size for random access,
    considering memory size
    """
    if wav_rxfilename.endswith("|"):
        # input piped command
        p = subprocess.Popen(wav_rxfilename[:-1], shell=True, stdout=subprocess.PIPE)
        data, samplerate = sf.load(io.BytesIO(p.stdout.read()), dtype="float32")
        # cannot seek
        data = data[start:end]
    elif wav_rxfilename == "-":
        # stdin
        data, samplerate = sf.load(sys.stdin, dtype="float32")
        # cannot seek
        data = data[start:end]
    else:
        # normal wav file
        data, samplerate = sf.load(wav_rxfilename, start=start, stop=end)
    return data, samplerate


def load_utt2spk(utt2spk_file):
    """returns dictionary { uttid: spkid }"""
    lines = [line.strip().split(None, 1) for line in open(utt2spk_file)]
    return {x[0]: x[1] for x in lines}


def load_spk2utt(spk2utt_file):
    """returns dictionary { spkid: list of uttids }"""
    if not os.path.exists(spk2utt_file):
        return None
    lines = [line.strip().split() for line in open(spk2utt_file)]
    return {x[0]: x[1:] for x in lines}


def load_reco2dur(reco2dur_file):
    """returns dictionary { recid: duration }"""
    if not os.path.exists(reco2dur_file):
        return None
    lines = [line.strip().split(None, 1) for line in open(reco2dur_file)]
    return {x[0]: float(x[1]) for x in lines}


def process_wav(wav_rxfilename, process):
    """This function returns preprocessed wav_rxfilename
    Args:
        wav_rxfilename: input
        process: command which can be connected via pipe,
                use stdin and stdout
    Returns:
        wav_rxfilename: output piped command
    """
    if wav_rxfilename.endswith("|"):
        # input piped command
        return wav_rxfilename + process + "|"
    else:
        # stdin "-" or normal file
        return "cat {} | {} |".format(wav_rxfilename, process)


def extract_segments(wavs, segments=None):
    """This function returns generator of segmented audio as
    (utterance id, numpy.float32 array)
    TODO?: sampling rate is not converted.
    """
    if segments is not None:
        # segments should be sorted by rec-id
        for seg in segments:
            wav = wavs[seg["rec"]]
            data, samplerate = load_wav(wav)
            st_sample = np.rint(seg["st"] * samplerate).astype(int)
            et_sample = np.rint(seg["et"] * samplerate).astype(int)
            yield seg["utt"], data[st_sample:et_sample]
    else:
        # segments file not found,
        # wav.scp is used as segmented audio list
        for rec in wavs:
            data, samplerate = load_wav(wavs[rec])
            yield rec, data


class KaldiData:
    def __init__(self, data_dir):
        self.data_dir = data_dir
        self.segments = load_segments_rechash(os.path.join(self.data_dir, "segments"))
        self.utt2spk = load_utt2spk(os.path.join(self.data_dir, "utt2spk"))
        self.wavs = load_wav_scp(os.path.join(self.data_dir, "wav.scp"))
        self.reco2dur = load_reco2dur(os.path.join(self.data_dir, "reco2dur"))
        self.spk2utt = load_spk2utt(os.path.join(self.data_dir, "spk2utt"))

    def load_wav(self, recid, start=0, end=None):
        data, rate = load_wav(self.wavs[recid], start, end)
        return data, rate



import cv2
from ultralytics import YOLO
import supervision as sv


def yolo_detector(model_path: str = "yolov8n.pt"):
    # 加载选择的合适的预训练模型
    model = YOLO(model_path)
    print("model-->:", model_path, type(model), model)
    model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
    print(f"have {model_million_params}M parameters")
    print("device-->:", model.device)

    # Define custom classes
    if hasattr(model, "set_classes"):
        model.set_classes(["person"])
        print(f"{model_path} set classes")
    # Run predictions
    # results = model('https://ultralytics.com/images/bus.jpg')

    # 打开视频流（0表示摄像头，或使用视频文件路径）
    cap = cv2.VideoCapture(0)

    while True:
        ret, frame = cap.read()
        if not ret:
            break
        print("cv read frame", type(frame))

        # 进行流式目标检测
        results = model(frame, stream=True)
        print("results-->:", type(results))
        is_check = False
        for item in results:
            print("item-->:", type(item), item)
            print("boxes-->:", item.boxes)
            detections = sv.Detections.from_ultralytics(item)
            print(
                "detections-->:",
                type(detections),
                len(detections),
                detections,
                detections.data["class_name"],
            )
            # 绘制检测结果
            labels = [
                f"{class_name} {confidence:.2f}"
                for class_name, confidence in zip(detections["class_name"], detections.confidence)
            ]

            BOUNDING_BOX_ANNOTATOR = sv.BoundingBoxAnnotator(thickness=2)
            LABEL_ANNOTATOR = sv.LabelAnnotator(
                text_thickness=2, text_scale=1, text_color=sv.Color.BLACK
            )
            annotated_image = frame.copy()
            annotated_image = BOUNDING_BOX_ANNOTATOR.annotate(annotated_image, detections)
            annotated_image = LABEL_ANNOTATOR.annotate(annotated_image, detections, labels=labels)
            annotated_frame = annotated_image
            # annotated_frame = item.plot()

            print("annotated_frame-->:", type(annotated_frame))

            # 显示结果
            cv2.imshow("YOLOv8 Stream", annotated_frame)
            cf_dict = {}
            for name in detections.data["class_name"]:
                cf_dict[name] = {"d_cn": 0, "d_cf": []}
            for t in zip(
                detections.xyxy,
                detections.confidence,
                detections.class_id,
                detections.data["class_name"],
            ):
                cf_dict[t[3]]["d_cn"] += 1
                cf_dict[t[3]]["d_cf"].append(t[1])

            if (
                "person" in cf_dict
                and max(cf_dict["person"]["d_cf"]) > 0.5
                and cf_dict["person"]["d_cn"] >= 1
            ):
                print("person detection-->:", cf_dict)
                is_check = True

        if is_check:
            # pass
            break

        if cv2.waitKey(1) & 0xFF == ord("q"):
            break

    cap.release()
    cv2.destroyAllWindows()


if __name__ == "__main__":
    # NOTE: just test to print all models stream predict info :)
    # yolo_detector("./models/yolov8n.pt")
    # yolo_detector("./models/yolov10n.pt")
    yolo_detector("./models/yolo11n.pt")
    # yolo_detector("./models/yolov8s-worldv2.pt")
    # yolo_detector("./models/rtdetr-l.pt")  # v1



from dataclasses import dataclass

import numpy as np
import torch
import torchaudio
from pytorch_lightning import LightningDataModule
from torch.utils.data import Dataset, DataLoader

import soundfile
# import librosa
import random

torch.set_num_threads(1)


@dataclass
class DataConfig:
    filelist_path: str
    sampling_rate: int
    num_samples: int
    batch_size: int
    num_workers: int

def collate_fn(batch):
    batch = [item for item in batch if item is not None]
    return torch.stack(batch, dim=0)

class VocosDataModule(LightningDataModule):
    def __init__(self, train_params: DataConfig, val_params: DataConfig):
        super().__init__()
        self.train_config = train_params
        self.val_config = val_params

    def _get_dataloder(self, cfg: DataConfig, train: bool):
        dataset = VocosDataset(cfg, train=train)
        dataloader = DataLoader(
            dataset, batch_size=cfg.batch_size, num_workers=cfg.num_workers, shuffle=train, pin_memory=True, collate_fn=collate_fn
        )
        return dataloader

    def train_dataloader(self) -> DataLoader:
        return self._get_dataloder(self.train_config, train=True)

    def val_dataloader(self) -> DataLoader:
        return self._get_dataloder(self.val_config, train=False)


class VocosDataset(Dataset):
    def __init__(self, cfg: DataConfig, train: bool):
        with open(cfg.filelist_path) as f:
            self.filelist = f.read().splitlines()
        self.sampling_rate = cfg.sampling_rate
        self.num_samples = cfg.num_samples
        self.train = train

    def __len__(self) -> int:
        return len(self.filelist)

    def __getitem__(self, index: int) -> torch.Tensor:
        audio_path = self.filelist[index]
        # y, sr = torchaudio.load(audio_path)
        # print(audio_path,"111")
        try:
            y1, sr = soundfile.read(audio_path)
            # y1, sr = librosa.load(audio_path,sr=None)
            y = torch.tensor(y1).float().unsqueeze(0)
            # if y.size(0) > 1:
            #     # mix to mono
            #     y = y.mean(dim=0, keepdim=True)
            if y.ndim > 2:
                # mix to mono
                # print("有问题哈,数据处理部分")
                # y = y.mean(dim=-1, keepdim=False)
                random_channel = random.randint(0, y.size(-1) - 1)
                y = y[:, :, random_channel] 

            gain = np.random.uniform(-1, -6) if self.train else -3
            y, _ = torchaudio.sox_effects.apply_effects_tensor(y, sr, [["norm", f"{gain:.2f}"]])
            if sr != self.sampling_rate:
                y = torchaudio.functional.resample(y, orig_freq=sr, new_freq=self.sampling_rate)
            if y.size(-1) < self.num_samples:
                pad_length = self.num_samples - y.size(-1)
                padding_tensor = y.repeat(1, 1 + pad_length // y.size(-1))
                y = torch.cat((y, padding_tensor[:, :pad_length]), dim=1)
            elif self.train:
                start = np.random.randint(low=0, high=y.size(-1) - self.num_samples + 1)
                y = y[:, start : start + self.num_samples]
            else:
                # During validation, take always the first segment for determinism
                y = y[:, : self.num_samples]

            return y[0]
        except Exception as e:
            print(f"Error processing file {audio_path} at index {index}: {e}")
            # 这里可以继续选择抛出异常，或者返回一个 None 表示无效数据
            return None

    # def __getitem__(self, index: int) -> torch.Tensor:
    #     audio_path = self.filelist[index]
    #     try:
    #         y, sr = torchaudio.load(audio_path)
    #         if y.size(0) > 1:
    #             # 随机选择一个通道
    #             random_channel = random.randint(0, y.size(0) - 1)
    #             y = y[random_channel, :].unsqueeze(0)  # 保持返回值为 (1, T) 的形式
    #         # gain = np.random.uniform(-1, -6) if self.train else -3
    #         # y, _ = torchaudio.sox_effects.apply_effects_tensor(y, sr, [["norm", f"{gain:.2f}"]])
    #         if sr != self.sampling_rate:
    #             y = torchaudio.functional.resample(y, orig_freq=sr, new_freq=self.sampling_rate)
    #         if y.size(-1) < self.num_samples:
    #             pad_length = self.num_samples - y.size(-1)
    #             padding_tensor = y.repeat(1, 1 + pad_length // y.size(-1))
    #             y = torch.cat((y, padding_tensor[:, :pad_length]), dim=1)
    #         elif self.train:
    #             start = np.random.randint(low=0, high=y.size(-1) - self.num_samples + 1)
    #             y = y[:, start: start + self.num_samples]
    #         else:
    #             # During validation, take always the first segment for determinism
    #             y = y[:, :self.num_samples]
    #         return y[0]
    #     except Exception as e:
    #         print(f"Error processing file {audio_path} at index {index}: {e}")
    #         # 这里可以继续选择抛出异常，或者返回一个 None 表示无效数据
    #         return None


import os
import logging
from typing import AsyncGenerator, Literal

try:
    import torch
    from diffusers import StableDiffusion3Pipeline, BitsAndBytesConfig, SD3Transformer2DModel
except ModuleNotFoundError as e:
    logging.error(
        "In order to use diffusers, you need to `pip install achatbot[diffusers,bitsandbytes]`."
        "Also, set `HF_API_KEY` environment variable."
    )
    raise Exception(f"Missing module: {e}")

from PIL import Image
from apipeline.frames.data_frames import Frame, ImageRawFrame
from apipeline.frames.sys_frames import ErrorFrame

from achatbot.processors.image.base import ImageGenProcessor


# https://huggingface.co/docs/api-inference/tasks/text-to-image


class HFStableDiffusionImageGenProcessor(ImageGenProcessor):
    def __init__(
        self,
        *,
        model: str = "stabilityai/stable-diffusion-3.5-large",
        width: int = 1024,
        height: int = 1024,
        steps: int = 28,
        guidance_scale: float = 7.5,  # like temperature
        is_quantizing: bool = False,
        negative_prompt: str = "monochrome, lowres, bad anatomy, worst quality, low quality",
        device: Literal["auto", "cpu", "cuda"] = "auto",
    ):
        super().__init__()
        self._width = width
        self._height = height
        self._steps = steps
        self._guidance_scale = guidance_scale
        self._negative_prompt = negative_prompt
        if is_quantizing:
            nf4_config = BitsAndBytesConfig(
                load_in_4bit=True, bnb_4bit_quant_type="nf4", bnb_4bit_compute_dtype=torch.bfloat16
            )
            model_nf4 = SD3Transformer2DModel.from_pretrained(
                model,
                subfolder="transformer",
                quantization_config=nf4_config,
                token=os.environ.get("HF_API_KEY"),
                torch_dtype=torch.bfloat16,
            )
            self._pipe = StableDiffusion3Pipeline.from_pretrained(
                model,
                token=os.environ.get("HF_API_KEY"),
                torch_dtype=torch.bfloat16,
                transformer=model_nf4,
            )
        else:
            self._pipe = StableDiffusion3Pipeline.from_pretrained(
                model,
                token=os.environ.get("HF_API_KEY"),
                torch_dtype=torch.bfloat16,
            )

        if device == "auto":
            self._pipe.enable_model_cpu_offload()
        else:
            self._pipe.to(device)
        # self._pipe.enable_xformers_memory_efficient_attention()
        logging.debug(f"sd pipeline: {self._pipe}, device: {self._pipe.device}")

    def set_aiohttp_session(self, session):
        pass

    async def run_image_gen(self, prompt: str) -> AsyncGenerator[Frame, None]:
        logging.debug(f"Generating image from prompt: {prompt}")

        image = self._pipe(
            prompt,
            num_inference_steps=self._steps,
            guidance_scale=self._guidance_scale,
            height=self._height,
            width=self._width,
            negative_prompt=self._negative_prompt,
            generator=torch.Generator(device=self._pipe.device).manual_seed(2),
        ).images[0]

        if not image:
            yield ErrorFrame("Image generation failed")
            return

        image = image.convert("RGB")
        # if image.size != (self._width, self._height):
        #    image = image.resize((self._width, self._height))
        frame = ImageRawFrame(
            image=image.tobytes(),
            size=image.size,
            format=image.format if image.format else "JPEG",
            mode=image.mode,
        )
        yield frame



from dataclasses import dataclass, field
from typing import List, Optional


from tensorrt_llm.llmapi.llm_utils import LlmArgs


from .sampling import LMGenerateArgs


@dataclass
class TensorRTLLMEngineArgs:
    """
    TensorRT-llm language model engine args
    """

    serv_args: dict = field(default_factory=lambda: LlmArgs().__dict__)
    gen_args: dict = field(default_factory=lambda: LMGenerateArgs().__dict__)


@dataclass
class TensorRTLLMRunnerArgs:
    """
    TensorRT-llm language model runner args
    """

    engine_dir: str
    max_output_len: Optional[int] = None
    lora_dir: Optional[List[str]] = None
    rank: int = 0
    debug_mode: bool = False
    lora_ckpt_source: str = "hf"
    medusa_choices: List[List[int]] = None
    # stream: torch.cuda.Stream = None
    gpu_weights_percent: float = 1
    enable_context_fmha_fp32_acc: Optional[bool] = None
    multi_block_mode: Optional[bool] = None


@dataclass
class TensorRTLLMRunnerEngineArgs:
    """
    TensorRT-llm language model runner engine args
    """

    serv_args: dict = field(default_factory=lambda: TensorRTLLMRunnerArgs().__dict__)
    gen_args: dict = field(default_factory=lambda: LMGenerateArgs().__dict__)



# Copyright [2023-11-28] <sxc19@mails.tsinghua.edu.cn, Xingchen Song>
#            2024 Alibaba Inc (authors: Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch

from cosyvoice.transformer.activation import Swish
from cosyvoice.transformer.subsampling import (
    LinearNoSubsampling,
    EmbedinigNoSubsampling,
    Conv1dSubsampling2,
    Conv2dSubsampling4,
    Conv2dSubsampling6,
    Conv2dSubsampling8,
)
from cosyvoice.transformer.embedding import (PositionalEncoding,
                                             RelPositionalEncoding,
                                             WhisperPositionalEncoding,
                                             LearnablePositionalEncoding,
                                             NoPositionalEncoding)
from cosyvoice.transformer.attention import (MultiHeadedAttention,
                                             RelPositionMultiHeadedAttention)
from cosyvoice.transformer.embedding import EspnetRelPositionalEncoding
from cosyvoice.transformer.subsampling import LegacyLinearNoSubsampling
from cosyvoice.llm.llm import TransformerLM, Qwen2LM
from cosyvoice.flow.flow import MaskedDiffWithXvec, CausalMaskedDiffWithXvec
from cosyvoice.hifigan.generator import HiFTGenerator
from cosyvoice.cli.model import CosyVoiceModel, CosyVoice2Model


COSYVOICE_ACTIVATION_CLASSES = {
    "hardtanh": torch.nn.Hardtanh,
    "tanh": torch.nn.Tanh,
    "relu": torch.nn.ReLU,
    "selu": torch.nn.SELU,
    "swish": getattr(torch.nn, "SiLU", Swish),
    "gelu": torch.nn.GELU,
}

COSYVOICE_SUBSAMPLE_CLASSES = {
    "linear": LinearNoSubsampling,
    "linear_legacy": LegacyLinearNoSubsampling,
    "embed": EmbedinigNoSubsampling,
    "conv1d2": Conv1dSubsampling2,
    "conv2d": Conv2dSubsampling4,
    "conv2d6": Conv2dSubsampling6,
    "conv2d8": Conv2dSubsampling8,
    'paraformer_dummy': torch.nn.Identity
}

COSYVOICE_EMB_CLASSES = {
    "embed": PositionalEncoding,
    "abs_pos": PositionalEncoding,
    "rel_pos": RelPositionalEncoding,
    "rel_pos_espnet": EspnetRelPositionalEncoding,
    "no_pos": NoPositionalEncoding,
    "abs_pos_whisper": WhisperPositionalEncoding,
    "embed_learnable_pe": LearnablePositionalEncoding,
}

COSYVOICE_ATTENTION_CLASSES = {
    "selfattn": MultiHeadedAttention,
    "rel_selfattn": RelPositionMultiHeadedAttention,
}


def get_model_type(configs):
    # NOTE CosyVoice2Model inherits CosyVoiceModel
    if isinstance(configs['llm'], TransformerLM) and isinstance(configs['flow'], MaskedDiffWithXvec) and isinstance(configs['hift'], HiFTGenerator):
        return CosyVoiceModel
    if isinstance(configs['llm'], Qwen2LM) and isinstance(configs['flow'], CausalMaskedDiffWithXvec) and isinstance(configs['hift'], HiFTGenerator):
        return CosyVoice2Model
    raise TypeError('No valid model type found!')



import os
from typing import Generator, List

from pydantic import BaseModel, Field
from dotenv import load_dotenv
import google.generativeai as genai
import instructor

from .. import types

# Load environment variables from .env file
load_dotenv(override=True)

genai.configure(api_key=os.environ["GOOGLE_API_KEY"])
client = instructor.from_gemini(
    client=genai.GenerativeModel(
        model_name="models/gemini-1.5-flash-latest",
    ),
    mode=instructor.Mode.GEMINI_JSON,
    generation_config={
        # "max_output_tokens": 1024,
        # "temperature": 1.0,
        # "top_p": 0.1,
        # "top_k": 40,
        # "response_mime_type": "text/plain",
    },
)


def extract_models(content: str, mode="text", **kwargs):
    match mode:
        case "partial":
            return extract_models_partial(content, **kwargs)
        case "iterable":
            return extract_models_iterable(content, **kwargs)
        case _:
            return extract_models_text(content, **kwargs)


def extract_models_partial(content: str, **kwargs):
    sys_prompt = get_system_prompt(**kwargs)
    res = client.create_partial(
        response_model=Chapters,
        messages=[
            {
                "role": "system",
                "content": sys_prompt,
            },
            {"role": "user", "content": content},
        ],
    )
    return res


def extract_models_iterable(content: str, **kwargs):
    sys_prompt = get_system_prompt(**kwargs)
    res = client.create_iterable(
        response_model=Chapters,
        messages=[
            {
                "role": "system",
                "content": sys_prompt,
            },
            {"role": "user", "content": content},
        ],
    )
    return res


def extract_models_text(content: str, **kwargs):
    sys_prompt = get_system_prompt(**kwargs)
    res = client.create(
        response_model=Chapters,
        messages=[
            {
                "role": "system",
                "content": sys_prompt,
            },
            {"role": "user", "content": content},
        ],
    )
    return res


class ChapterSystemPromptArgs(BaseModel):
    language: str = "en"


def get_system_prompt(**kwargs) -> str:
    r"""
    !NOTE: the same as ell use python function  :)
    """
    args = ChapterSystemPromptArgs(**kwargs)
    return f"Analyze the given YouTube transcript and extract chapters. For each chapter, provide a start timestamp, end timestamp, title, and summary. Output language should be in {types.TO_LLM_LANGUAGE[args.language]}"


class Chapter(BaseModel):
    start_ts: float = Field(
        ...,
        description="The start timestamp indicating when the chapter starts in the video.",
    )
    end_ts: float = Field(
        ...,
        description="The end timestamp indicating when the chapter ends in the video.",
    )
    title: str = Field(..., description="A concise and descriptive title for the chapter.")
    summary: str = Field(
        ...,
        description="A brief summary of the chapter's content, don't use words like 'the speaker'",
    )


class Chapters(BaseModel):
    chapters: list[Chapter]


def console_table(chapters: Generator[Chapters, None, None] | List[Chapter]):
    from rich.table import Table
    from rich.live import Live

    table = Table(title="Chapters")
    table.add_column("Title", style="magenta")
    table.add_column("Description", style="green")
    table.add_column("Start", style="cyan")
    table.add_column("End", style="cyan")

    with Live(refresh_per_second=4) as live:
        for extraction in chapters:
            if not extraction.chapters:
                continue

            new_table = Table(title="Video Chapters")
            new_table.add_column("Title", style="magenta")
            new_table.add_column("Description", style="green")
            new_table.add_column("Start", style="cyan")
            new_table.add_column("End", style="cyan")

            for chapter in extraction.chapters:
                new_table.add_row(
                    chapter.title,
                    chapter.summary,
                    f"{chapter.start_ts:.2f}" if chapter.start_ts else "",
                    f"{chapter.end_ts:.2f}" if chapter.end_ts else "",
                )
                new_table.add_row("", "", "", "")  # Add an empty row for spacing

            live.update(new_table)



"""
From https://github.com/lm-sys/FastChat/blob/main/fastchat/conversation.py
"""

import dataclasses
from enum import IntEnum, auto
from typing import Any, Dict, List


class SeparatorStyle(IntEnum):
    """Separator styles."""

    DeepSeek = auto()
    DeepSeekV2 = auto()
    PLAIN = auto()
    ALIGNMENT = auto()


@dataclasses.dataclass
class Conversation:
    """A class that manages prompt templates and keeps all conversation history."""

    # The name of this template
    name: str
    # The template of the system prompt
    system_template: str = "{system_message}"
    # The system message
    system_message: str = ""
    # The names of two roles
    roles: List[str] = (("USER", "ASSISTANT"),)
    # All messages. Each item is (role, message).
    messages: List[List[str]] = ()
    # The number of few shot examples
    offset: int = 0
    # The separator style and configurations
    sep_style: SeparatorStyle = SeparatorStyle.DeepSeek
    sep: str = "\n"
    sep2: str = None
    # Stop criteria (the default one is EOS token)
    stop_str: str = None
    # Stops generation if meeting any token in this list
    stop_token_ids: List[int] = None

    def get_prompt(self) -> str:
        """Get the prompt for generation."""
        system_prompt = self.system_template.format(system_message=self.system_message)
        if self.sep_style == SeparatorStyle.DeepSeek:
            seps = [self.sep, self.sep2]
            if system_prompt == "" or system_prompt is None:
                ret = ""
            else:
                ret = system_prompt + seps[0]
            for i, (role, message) in enumerate(self.messages):
                if message:
                    ret += role + ": " + message + seps[i % 2]
                else:
                    ret += role + ":"
            return ret
        elif self.sep_style == SeparatorStyle.DeepSeekV2:
            seps = [self.sep, self.sep2]
            if system_prompt == "" or system_prompt is None:
                ret = ""
            else:
                ret = system_prompt + seps[0]
            for i, (role, message) in enumerate(self.messages):
                if message:
                    if role == "User":
                        ret += "<｜sft▁begin｜>\n" + message + self.sep #<｜sft▁begin｜>User Input<｜sft▁end｜>\nResponse<｜end▁of▁sentence｜>
                    else:
                        ret += message + self.sep2
                else:
                    ret = ret
            return ret

        elif self.sep_style == SeparatorStyle.PLAIN:
            seps = [self.sep, self.sep2]
            ret = ""
            for i, (role, message) in enumerate(self.messages):
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    if i % 2 == 0:
                        ret += message + seps[i % 2]
                    else:
                        ret += message + seps[i % 2]
                else:
                    ret += ""
            return ret
        elif self.sep_style == SeparatorStyle.ALIGNMENT:
            seps = [self.sep, self.sep2]
            ret = ""
            for i, (role, message) in enumerate(self.messages):
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    if i % 2 == 0:
                        ret += '<image>\n' + seps[i % 2]
                    else:
                        ret += message + seps[i % 2]
                else:
                    ret += ""
            return ret
        else:
            raise ValueError(f"Invalid style: {self.sep_style}")

    def set_system_message(self, system_message: str):
        """Set the system message."""
        self.system_message = system_message

    def append_message(self, role: str, message: str):
        """Append a new message."""
        self.messages.append([role, message])

    def update_last_message(self, message: str):
        """Update the last output.

        The last message is typically set to be None when constructing the prompt,
        so we need to update it in-place after getting the response from a model.
        """
        self.messages[-1][1] = message

    def reset_message(self):
        """Reset a new message."""
        self.messages = []

    def to_gradio_chatbot(self):
        """Convert the conversation to gradio chatbot format."""
        ret = []
        for i, (role, msg) in enumerate(self.messages[self.offset :]):
            if i % 2 == 0:
                ret.append([msg, None])
            else:
                ret[-1][-1] = msg
        return ret

    def to_openai_api_messages(self):
        """Convert the conversation to OpenAI chat completion format."""
        system_prompt = self.system_template.format(system_message=self.system_message)
        ret = [{"role": "system", "content": system_prompt}]

        for i, (_, msg) in enumerate(self.messages[self.offset :]):
            if i % 2 == 0:
                ret.append({"role": "user", "content": msg})
            else:
                if msg is not None:
                    ret.append({"role": "assistant", "content": msg})
        return ret

    def copy(self):
        return Conversation(
            name=self.name,
            system_template=self.system_template,
            system_message=self.system_message,
            roles=self.roles,
            messages=[[x, y] for x, y in self.messages],
            offset=self.offset,
            sep_style=self.sep_style,
            sep=self.sep,
            sep2=self.sep2,
            stop_str=self.stop_str,
            stop_token_ids=self.stop_token_ids,
        )

    def dict(self):
        return {
            "template_name": self.name,
            "system_message": self.system_message,
            "roles": self.roles,
            "messages": self.messages,
            "offset": self.offset,
        }


# A global registry for all conversation templates
conv_templates: Dict[str, Conversation] = {}


def register_conv_template(template: Conversation, override: bool = False):
    """Register a new conversation template."""
    if not override:
        assert template.name not in conv_templates, f"{template.name} has been registered."

    conv_templates[template.name] = template


def get_conv_template(name: str) -> Conversation:
    """Get a conversation template."""
    return conv_templates[name].copy()


# register_conv_template(
#     Conversation(
#         name="deepseek",
#         system_template="{system_message}",
#         # system_message="You are a helpful assistant. Please answer truthfully and write out your "
#         # "thinking step by step to be sure you get the right answer.",
#         system_message="",
#         roles=("User", "Assistant"),
#         messages=(),
#         offset=0,
#         sep_style=SeparatorStyle.DeepSeek,
#         sep="\n\n",
#         sep2="<｜end▁of▁sentence｜>",
#         stop_token_ids=[100001],
#         stop_str=["User:", "<｜end▁of▁sentence｜>"]
#     )
# )
register_conv_template(
    Conversation(
        name="deepseek",
        system_template="{system_message}",
        # system_message="You are a helpful assistant. Please answer truthfully and write out your "
        # "thinking step by step to be sure you get the right answer.",
        system_message="",
        roles=("<|User|>", "<|Assistant|>"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.DeepSeek,
        sep="\n\n",
        sep2="<｜end▁of▁sentence｜>",
        stop_token_ids=[100001],
        stop_str=["User:", "<｜end▁of▁sentence｜>"]
    )
)
# register_conv_template(
#     Conversation(
#         name="deepseekv2",
#         system_template="{system_message}",
#         system_message="",
#         roles=("User", "Assistant"),
#         messages=(),
#         offset=0,
#         sep_style=SeparatorStyle.DeepSeekV2,
#         sep="\n<｜sft▁end｜>",
#         sep2="<｜end▁of▁sentence｜>",
#         stop_token_ids=[100001],
#         stop_str=["User:", "<｜end▁of▁sentence｜>"]
#     )
# )
register_conv_template(
    Conversation(
        name="deepseekv2",
        system_template="{system_message}",
        system_message="",
        roles=("|<User>|", "|<Assistant>|"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.DeepSeekV2,
        sep="\n<｜sft▁end｜>",
        sep2="<｜end▁of▁sentence｜>",
        stop_token_ids=[100001],
        stop_str=["User:", "<｜end▁of▁sentence｜>"]
    )
)


register_conv_template(
    Conversation(
        name="plain",
        system_template="",
        system_message="",
        roles=("", ""),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.PLAIN,
        sep="",
        sep2="",
        stop_token_ids=[100001],
        stop_str=['</s>'],
    )
)


register_conv_template(
    Conversation(
        name="alignment",
        system_template="",
        system_message="",
        roles=("", ""),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ALIGNMENT,
        sep="",
        sep2="",
        stop_token_ids=[100001],
        stop_str=['</s>'],
    )
)


if __name__ == "__main__":
    print("deepseek template:")
    conv = get_conv_template("deepseek")
    conv.append_message(conv.roles[0], "Hello!")
    conv.append_message(conv.roles[1], "Hi! This is Tony.")
    conv.append_message(conv.roles[0], "Who are you?")
    conv.append_message(conv.roles[1], "I am a helpful assistant.")
    conv.append_message(conv.roles[0], "How are you?")
    conv.append_message(conv.roles[1], None)
    print(conv.get_prompt())

    print("deepseekv2 template:")
    conv = get_conv_template("deepseekv2")
    conv.append_message(conv.roles[0], "Hello!")
    conv.append_message(conv.roles[1], "Hi! This is Tony.")
    conv.append_message(conv.roles[0], "Who are you?")
    conv.append_message(conv.roles[1], "I am a helpful assistant.")
    conv.append_message(conv.roles[0], "How are you?")
    conv.append_message(conv.roles[1], None)
    print(conv.get_prompt())



import logging
import json
import os
import threading
import queue

import grpc
import uuid

from achatbot.common.grpc.idl.tts_pb2_grpc import TTSServicer
from achatbot.common.grpc.idl.tts_pb2 import (
    GetStreamInfoReponse,
    GetStreamInfoRequest,
    GetVoicesRequest,
    GetVoicesResponse,
    LoadModelRequest,
    LoadModelResponse,
    SetVoiceRequest,
    SetVoiceResponse,
    SynthesizeRequest,
    SynthesizeResponse,
)
from achatbot.common.factory import EngineClass
from achatbot.common.session import Session
from achatbot.common.types import SessionCtx, ITts
from achatbot.modules.speech.tts import TTSEnvInit


def get_session_id(context: grpc.ServicerContext):
    for item in context.invocation_metadata():
        if item.key == "client_id":
            return item.value
    return str(uuid.uuid4())


class TTS(TTSServicer):
    def __init__(self) -> None:
        super().__init__()
        self._tts_instance: EngineClass | ITts = None
        self._is_initializing = False
        self._lock = threading.Lock()
        self.init_queue = queue.Queue()

    def get_tts_instance(self):
        """获取当前的 TTS 实例"""
        with self._lock:
            return self._tts_instance

    def LoadModel(self, request: LoadModelRequest, context: grpc.ServicerContext):
        logging.debug(f"LoadModel request: {request}")
        kwargs = json.loads(request.json_kwargs)
        logging.debug(f"LoadModel kwargs: {kwargs}")

        # print("--->", self.get_tts_instance(), request.is_reload)
        if self.get_tts_instance() and request.is_reload is False:
            logging.info("TTS is already initializing un reload, ignoring this signal.")
            return LoadModelResponse()

        with self._lock:
            if self._is_initializing:
                logging.info("TTS is already initializing, ignoring this signal.")
                return LoadModelResponse()

        init_event = threading.Event()
        init_result = {"success": False, "error": None}

        init_signal = {
            "tts_tag": request.tts_tag,
            "kwargs": kwargs,
            "event": init_event,
            "result": init_result,
        }
        self.init_queue.put(init_signal)
        # logging.info(f"Initialization signal added to queue: {init_signal}")

        if not init_event.wait(timeout=int(os.getenv("INIT_TIMEOUT", "600"))):
            # default wait 10 minutes
            logging.error("Initialization timed out.")
            raise Exception("TTS initialization timed out.")

        if not init_result["success"]:
            logging.error(f"Initialization failed: {init_result['error']}")
            raise Exception(f"TTS initialization failed: {init_result['error']}")

        logging.info("TTS initialization completed successfully.")
        return LoadModelResponse()

    def GetVoices(self, request: GetVoicesRequest, context: grpc.ServicerContext):
        voices = self.get_tts_instance().get_voices()
        return GetVoicesResponse(voices=voices)

    def SetVoice(self, request: SetVoiceRequest, context: grpc.ServicerContext):
        self.get_tts_instance().set_voice(request.voice)
        return SetVoiceResponse()

    def GetStreamInfo(self, request: GetStreamInfoRequest, context: grpc.ServicerContext):
        info = self.get_tts_instance().get_stream_info()
        return GetStreamInfoReponse(
            format=info["format"],
            rate=info["rate"],
            channels=info["channels"],
            sample_width=info["sample_width"],
        )

    def SynthesizeUS(self, request: SynthesizeRequest, context: grpc.ServicerContext):
        logging.debug(f"SynthesizeUS request: {request} json.kwargs:{request.json_kwargs}")
        session = Session(**SessionCtx(get_session_id(context)).__dict__)
        if request.json_kwargs:
            kwargs = json.loads(request.json_kwargs)
            logging.debug(f"SynthesizeUS kwargs: {kwargs}")
            session.ctx.state = kwargs
        session.ctx.state["tts_text"] = request.tts_text
        iter = self.get_tts_instance().synthesize_sync(session)
        for i, chunk in enumerate(iter):
            logging.debug(f"get {i} chunk {len(chunk)}")
            yield SynthesizeResponse(tts_audio=chunk)


def main_thread_init(tts_service: TTS):
    logging.info("Main thread initializing TTS engine.")
    while True:
        try:
            init_signal = tts_service.init_queue.get()
            logging.debug(f"Main thread received initialization signal: {init_signal}")
            tts_tag = init_signal["tts_tag"]
            kwargs = init_signal["kwargs"]
            init_event = init_signal["event"]
            init_result = init_signal["result"]

            with tts_service._lock:
                if tts_service._is_initializing:
                    logging.info("Duplicate initialization signal detected, skipping.")
                    continue
                tts_service._is_initializing = True

            try:
                tts_instance = TTSEnvInit.getEngine(tts_tag, **kwargs)
                logging.info(f"Main thread initialized TTS engine: {tts_instance.TAG}")
                with tts_service._lock:
                    tts_service._tts_instance = tts_instance
                init_result["success"] = True
            except Exception as e:
                logging.error(f"Failed to initialize TTS engine: {e}")
                init_result["success"] = False
                init_result["error"] = str(e)
            finally:
                with tts_service._lock:
                    tts_service._is_initializing = False

            init_event.set()

        except Exception as e:
            logging.error(f"Error in main thread initialization loop: {e}")



import logging
import time

import grpc

from demo.grpc.idl.echo_pb2 import EchoRequest
from demo.grpc.idl.echo_pb2_grpc import EchoStub
from demo.grpc.interceptors.authentication_client import add_authentication


def echo_uu(channel):
    print("What is your name?")
    name = input()
    echo_stub = EchoStub(channel)
    request_data = EchoRequest(name=name)
    response = echo_stub.EchoUU(request_data)
    print(response.echo)


def echo_su(channel):
    echo_stub = EchoStub(channel)
    request_iterator = iter([EchoRequest(name="xiaohei"), EchoRequest(name="xiaowu")])
    response = echo_stub.EchoSU(request_iterator)
    print(response.echo)


def echo_us(channel):
    echo_stub = EchoStub(channel)
    request_data = EchoRequest(name="echo_us")
    response_iterator = echo_stub.EchoUS(request_data)
    for response in response_iterator:
        print(response.echo)


def echo_iter():
    cn = 1
    while True:
        yield EchoRequest(name=f"client_echo_iter{cn}")
        cn += 1
        time.sleep(1)


def echo_ss(channel):
    echo_stub = EchoStub(channel)
    response_iterator = echo_stub.EchoSS(echo_iter())
    for response in response_iterator:
        print(response.echo)


logging.basicConfig(level=logging.DEBUG)

# python -m demo.grpc.echo.client 2> err_cli.lo
if __name__ == "__main__":
    try:
        # todo: up to the rpc gateway to auth
        token = "oligei"
        authentication = add_authentication("authorization", token)
        channel = grpc.insecure_channel("localhost:50051")
        channel = grpc.intercept_channel(channel, authentication)
        for op in [echo_uu, echo_su, echo_us, echo_ss]:
            print(f"--{op}--")
            try:
                op(channel)
            except grpc.RpcError as e:
                logging.error(e)
    except Exception as e:
        logging.error(f"Exception: {e}")
    finally:
        channel and channel.close()



import os
import subprocess
import sys

#!/usr/bin/env python


def main():
    # Make path relative to this file
    script_path = os.path.join(
        os.path.dirname(__file__), "../../fish_speech/models/vqgan/inference.py"
    )
    subprocess.run(["python", script_path] + sys.argv[1:])


if __name__ == "__main__":
    main()



""" from https://github.com/keithito/tacotron """

'''
Cleaners are transformations that run over the input text at both training and eval time.

Cleaners can be selected by passing a comma-delimited list of cleaner names as the "cleaners"
hyperparameter. Some cleaners are English-specific. You'll typically want to use:
  1. "english_cleaners" for English text
  2. "transliteration_cleaners" for non-English text that can be transliterated to ASCII using
     the Unidecode library (https://pypi.python.org/pypi/Unidecode)
  3. "basic_cleaners" if you do not want to transliterate (in this case, you should also update
     the symbols in symbols.py to match your data).
'''


# Regular expression matching whitespace:


import re
import inflect
from unidecode import unidecode
import eng_to_ipa as ipa
_inflect = inflect.engine()
_comma_number_re = re.compile(r'([0-9][0-9\,]+[0-9])')
_decimal_number_re = re.compile(r'([0-9]+\.[0-9]+)')
_pounds_re = re.compile(r'£([0-9\,]*[0-9]+)')
_dollars_re = re.compile(r'\$([0-9\.\,]*[0-9]+)')
_ordinal_re = re.compile(r'[0-9]+(st|nd|rd|th)')
_number_re = re.compile(r'[0-9]+')

# List of (regular expression, replacement) pairs for abbreviations:
_abbreviations = [(re.compile('\\b%s\\.' % x[0], re.IGNORECASE), x[1]) for x in [
    ('mrs', 'misess'),
    ('mr', 'mister'),
    ('dr', 'doctor'),
    ('st', 'saint'),
    ('co', 'company'),
    ('jr', 'junior'),
    ('maj', 'major'),
    ('gen', 'general'),
    ('drs', 'doctors'),
    ('rev', 'reverend'),
    ('lt', 'lieutenant'),
    ('hon', 'honorable'),
    ('sgt', 'sergeant'),
    ('capt', 'captain'),
    ('esq', 'esquire'),
    ('ltd', 'limited'),
    ('col', 'colonel'),
    ('ft', 'fort'),
]]


# List of (ipa, lazy ipa) pairs:
_lazy_ipa = [(re.compile('%s' % x[0]), x[1]) for x in [
    ('r', 'ɹ'),
    ('æ', 'e'),
    ('ɑ', 'a'),
    ('ɔ', 'o'),
    ('ð', 'z'),
    ('θ', 's'),
    ('ɛ', 'e'),
    ('ɪ', 'i'),
    ('ʊ', 'u'),
    ('ʒ', 'ʥ'),
    ('ʤ', 'ʥ'),
    ('ˈ', '↓'),
]]

# List of (ipa, lazy ipa2) pairs:
_lazy_ipa2 = [(re.compile('%s' % x[0]), x[1]) for x in [
    ('r', 'ɹ'),
    ('ð', 'z'),
    ('θ', 's'),
    ('ʒ', 'ʑ'),
    ('ʤ', 'dʑ'),
    ('ˈ', '↓'),
]]

# List of (ipa, ipa2) pairs
_ipa_to_ipa2 = [(re.compile('%s' % x[0]), x[1]) for x in [
    ('r', 'ɹ'),
    ('ʤ', 'dʒ'),
    ('ʧ', 'tʃ')
]]


def expand_abbreviations(text):
    for regex, replacement in _abbreviations:
        text = re.sub(regex, replacement, text)
    return text


def collapse_whitespace(text):
    return re.sub(r'\s+', ' ', text)


def _remove_commas(m):
    return m.group(1).replace(',', '')


def _expand_decimal_point(m):
    return m.group(1).replace('.', ' point ')


def _expand_dollars(m):
    match = m.group(1)
    parts = match.split('.')
    if len(parts) > 2:
        return match + ' dollars'  # Unexpected format
    dollars = int(parts[0]) if parts[0] else 0
    cents = int(parts[1]) if len(parts) > 1 and parts[1] else 0
    if dollars and cents:
        dollar_unit = 'dollar' if dollars == 1 else 'dollars'
        cent_unit = 'cent' if cents == 1 else 'cents'
        return '%s %s, %s %s' % (dollars, dollar_unit, cents, cent_unit)
    elif dollars:
        dollar_unit = 'dollar' if dollars == 1 else 'dollars'
        return '%s %s' % (dollars, dollar_unit)
    elif cents:
        cent_unit = 'cent' if cents == 1 else 'cents'
        return '%s %s' % (cents, cent_unit)
    else:
        return 'zero dollars'


def _expand_ordinal(m):
    return _inflect.number_to_words(m.group(0))


def _expand_number(m):
    num = int(m.group(0))
    if num > 1000 and num < 3000:
        if num == 2000:
            return 'two thousand'
        elif num > 2000 and num < 2010:
            return 'two thousand ' + _inflect.number_to_words(num % 100)
        elif num % 100 == 0:
            return _inflect.number_to_words(num // 100) + ' hundred'
        else:
            return _inflect.number_to_words(num, andword='', zero='oh', group=2).replace(', ', ' ')
    else:
        return _inflect.number_to_words(num, andword='')


def normalize_numbers(text):
    text = re.sub(_comma_number_re, _remove_commas, text)
    text = re.sub(_pounds_re, r'\1 pounds', text)
    text = re.sub(_dollars_re, _expand_dollars, text)
    text = re.sub(_decimal_number_re, _expand_decimal_point, text)
    text = re.sub(_ordinal_re, _expand_ordinal, text)
    text = re.sub(_number_re, _expand_number, text)
    return text


def mark_dark_l(text):
    return re.sub(r'l([^aeiouæɑɔəɛɪʊ ]*(?: |$))', lambda x: 'ɫ'+x.group(1), text)


def english_to_ipa(text):
    text = unidecode(text).lower()
    text = expand_abbreviations(text)
    text = normalize_numbers(text)
    phonemes = ipa.convert(text)
    phonemes = collapse_whitespace(phonemes)
    return phonemes


def english_to_lazy_ipa(text):
    text = english_to_ipa(text)
    for regex, replacement in _lazy_ipa:
        text = re.sub(regex, replacement, text)
    return text


def english_to_ipa2(text):
    text = english_to_ipa(text)
    text = mark_dark_l(text)
    for regex, replacement in _ipa_to_ipa2:
        text = re.sub(regex, replacement, text)
    return text.replace('...', '…')


def english_to_lazy_ipa2(text):
    text = english_to_ipa(text)
    for regex, replacement in _lazy_ipa2:
        text = re.sub(regex, replacement, text)
    return text



import random
import sys
from importlib.resources import files

import soundfile as sf
import tqdm
from cached_path import cached_path

from f5_tts.infer.utils_infer import (
    hop_length,
    infer_process,
    load_model,
    load_vocoder,
    preprocess_ref_audio_text,
    remove_silence_for_generated_wav,
    save_spectrogram,
    transcribe,
    target_sample_rate,
)
from f5_tts.model import DiT, UNetT
from f5_tts.model.utils import seed_everything


class F5TTS:
    def __init__(
        self,
        model_type="F5-TTS",
        ckpt_file="",
        vocab_file="",
        ode_method="euler",
        use_ema=True,
        vocoder_name="vocos",
        local_path=None,
        device=None,
        hf_cache_dir=None,
    ):
        # Initialize parameters
        self.final_wave = None
        self.target_sample_rate = target_sample_rate
        self.hop_length = hop_length
        self.seed = -1
        self.mel_spec_type = vocoder_name

        # Set device
        if device is not None:
            self.device = device
        else:
            import torch

            self.device = "cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu"

        # Load models
        self.load_vocoder_model(vocoder_name, local_path=local_path, hf_cache_dir=hf_cache_dir)
        self.load_ema_model(
            model_type, ckpt_file, vocoder_name, vocab_file, ode_method, use_ema, hf_cache_dir=hf_cache_dir
        )

    def load_vocoder_model(self, vocoder_name, local_path=None, hf_cache_dir=None):
        self.vocoder = load_vocoder(vocoder_name, local_path is not None, local_path, self.device, hf_cache_dir)

    def load_ema_model(self, model_type, ckpt_file, mel_spec_type, vocab_file, ode_method, use_ema, hf_cache_dir=None):
        if model_type == "F5-TTS":
            if not ckpt_file:
                if mel_spec_type == "vocos":
                    ckpt_file = str(
                        cached_path("hf://SWivid/F5-TTS/F5TTS_Base/model_1200000.safetensors", cache_dir=hf_cache_dir)
                    )
                elif mel_spec_type == "bigvgan":
                    ckpt_file = str(
                        cached_path("hf://SWivid/F5-TTS/F5TTS_Base_bigvgan/model_1250000.pt", cache_dir=hf_cache_dir)
                    )
            model_cfg = dict(dim=1024, depth=22, heads=16, ff_mult=2, text_dim=512, conv_layers=4)
            model_cls = DiT
        elif model_type == "E2-TTS":
            if not ckpt_file:
                ckpt_file = str(
                    cached_path("hf://SWivid/E2-TTS/E2TTS_Base/model_1200000.safetensors", cache_dir=hf_cache_dir)
                )
            model_cfg = dict(dim=1024, depth=24, heads=16, ff_mult=4)
            model_cls = UNetT
        else:
            raise ValueError(f"Unknown model type: {model_type}")

        self.ema_model = load_model(
            model_cls, model_cfg, ckpt_file, mel_spec_type, vocab_file, ode_method, use_ema, self.device
        )

    def transcribe(self, ref_audio, language=None):
        return transcribe(ref_audio, language)

    def export_wav(self, wav, file_wave, remove_silence=False):
        sf.write(file_wave, wav, self.target_sample_rate)

        if remove_silence:
            remove_silence_for_generated_wav(file_wave)

    def export_spectrogram(self, spect, file_spect):
        save_spectrogram(spect, file_spect)

    def infer(
        self,
        ref_file,
        ref_text,
        gen_text,
        show_info=print,
        progress=tqdm,
        target_rms=0.1,
        cross_fade_duration=0.15,
        sway_sampling_coef=-1,
        cfg_strength=2,
        nfe_step=32,
        speed=1.0,
        fix_duration=None,
        remove_silence=False,
        file_wave=None,
        file_spect=None,
        seed=-1,
    ):
        if seed == -1:
            seed = random.randint(0, sys.maxsize)
        seed_everything(seed)
        self.seed = seed

        ref_file, ref_text = preprocess_ref_audio_text(ref_file, ref_text, device=self.device)

        wav, sr, spect = infer_process(
            ref_file,
            ref_text,
            gen_text,
            self.ema_model,
            self.vocoder,
            self.mel_spec_type,
            show_info=show_info,
            progress=progress,
            target_rms=target_rms,
            cross_fade_duration=cross_fade_duration,
            nfe_step=nfe_step,
            cfg_strength=cfg_strength,
            sway_sampling_coef=sway_sampling_coef,
            speed=speed,
            fix_duration=fix_duration,
            device=self.device,
        )

        if file_wave is not None:
            self.export_wav(wav, file_wave, remove_silence)

        if file_spect is not None:
            self.export_spectrogram(spect, file_spect)

        return wav, sr, spect


if __name__ == "__main__":
    f5tts = F5TTS()

    wav, sr, spect = f5tts.infer(
        ref_file=str(files("f5_tts").joinpath("infer/examples/basic/basic_ref_en.wav")),
        ref_text="some call me nature, others call me mother nature.",
        gen_text="""I don't really care what you call me. I've been a silent spectator, watching species evolve, empires rise and fall. But always remember, I am mighty and enduring. Respect me and I'll nurture you; ignore me and you shall face the consequences.""",
        file_wave=str(files("f5_tts").joinpath("../../tests/api_out.wav")),
        file_spect=str(files("f5_tts").joinpath("../../tests/api_out.png")),
        seed=-1,  # random seed = -1
    )

    print("seed :", f5tts.seed)



import ast
import glob
import json
from collections import OrderedDict
from pathlib import Path

from loguru import logger

from .core import DEFAULT_LANGUAGE, I18N_FILE_PATH


def extract_i18n_strings(node):
    i18n_strings = []

    if (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Name)
        and node.func.id == "i18n"
    ):
        for arg in node.args:
            if isinstance(arg, ast.Str):
                i18n_strings.append(arg.s)

    for child_node in ast.iter_child_nodes(node):
        i18n_strings.extend(extract_i18n_strings(child_node))

    return i18n_strings


# scan the directory for all .py files (recursively)
# for each file, parse the code into an AST
# for each AST, extract the i18n strings

strings = []
folders = ["fish_speech", "tools"]
# for filename in glob.iglob("**/*.py", recursive=True):
for folder in folders:
    for f in Path(folder).rglob("*.py"):
        code = f.read_text(encoding="utf-8")
        if "i18n(" in code:
            tree = ast.parse(code)
            i18n_strings = extract_i18n_strings(tree)
            logger.info(f"Found {len(i18n_strings)} i18n strings in {f}")
            strings.extend(i18n_strings)

code_keys = set(strings)
logger.info(f"Total unique: {len(code_keys)}")


standard_file = I18N_FILE_PATH / f"{DEFAULT_LANGUAGE}.json"
with open(standard_file, "r", encoding="utf-8") as f:
    standard_data = json.load(f, object_pairs_hook=OrderedDict)
standard_keys = set(standard_data.keys())

# Define the standard file name
unused_keys = standard_keys - code_keys
logger.info(f"Found {len(unused_keys)} unused keys in {standard_file}")
for unused_key in unused_keys:
    logger.info(f"\t{unused_key}")

missing_keys = code_keys - standard_keys
logger.info(f"Found {len(missing_keys)} missing keys in {standard_file}")
for missing_key in missing_keys:
    logger.info(f"\t{missing_key}")

code_keys_dict = OrderedDict()
for s in strings:
    code_keys_dict[s] = s

# write back
with open(standard_file, "w", encoding="utf-8") as f:
    json.dump(code_keys_dict, f, ensure_ascii=False, indent=4, sort_keys=True)
    f.write("\n")

logger.info(f"Updated {standard_file}")


# Define the standard file name
standard_file = I18N_FILE_PATH / f"{DEFAULT_LANGUAGE}.json"

# Find all JSON files in the directory
dir_path = I18N_FILE_PATH
languages = [f for f in dir_path.glob("*.json") if f.stem != DEFAULT_LANGUAGE]

# Load the standard file
with open(standard_file, "r", encoding="utf-8") as f:
    standard_data = json.load(f, object_pairs_hook=OrderedDict)

# Loop through each language file
for lang_file in languages:
    # Load the language file
    with open(lang_file, "r", encoding="utf-8") as f:
        lang_data = json.load(f, object_pairs_hook=OrderedDict)

    # Find the difference between the language file and the standard file
    diff = set(standard_data.keys()) - set(lang_data.keys())

    miss = set(lang_data.keys()) - set(standard_data.keys())

    # Add any missing keys to the language file
    for key in diff:
        lang_data[key] = "#!" + key
        logger.info(f"Added missing key: {key} to {lang_file}")

    # Del any extra keys to the language file
    for key in miss:
        del lang_data[key]
        logger.info(f"Del extra key: {key} from {lang_file}")

    # Sort the keys of the language file to match the order of the standard file
    lang_data = OrderedDict(
        sorted(lang_data.items(), key=lambda x: list(standard_data.keys()).index(x[0]))
    )

    # Save the updated language file
    with open(lang_file, "w", encoding="utf-8") as f:
        json.dump(lang_data, f, ensure_ascii=False, indent=4, sort_keys=True)
        f.write("\n")

    logger.info(f"Updated {lang_file}")

logger.info("Done")



from dataclasses import dataclass, field


@dataclass
class LMGenerateArgs:
    """lm generation sampling arguments"""

    lm_gen_seed: int = field(
        default=42,
        metadata={"help": "The seed to use for the language model. Default is 42."},
    )
    lm_max_length: int = field(
        default=2048,
        metadata={
            "help": "Corresponds to the length of the input prompt + max_new_tokens. Its effect is overridden by max_new_tokens, if also set. Default is 2048."
        },
    )
    lm_gen_max_new_tokens: int = field(
        default=1024,
        metadata={
            "help": "Maximum number of new tokens to generate in a single completion. Default is 1024."
        },
    )
    lm_gen_min_new_tokens: int = field(
        default=1,
        metadata={
            "help": "Minimum number of new tokens to generate in a single completion. Default is 1."
        },
    )
    lm_gen_do_sample: bool = field(
        default=True,
        metadata={
            "help": "Whether to use sampling; set this to False for deterministic outputs. Default is False."
        },
    )
    lm_gen_temperature: float = field(
        default=0.1,
        metadata={
            "help": "Controls the randomness of the output. Set to 0.0 for deterministic (repeatable) outputs. Default is 0.1."
        },
    )
    lm_gen_top_k: int = field(
        default=20,
        metadata={
            "help": "Changing the top - k parameter sets the size of the shortlist the model samples from as it outputs each token. Setting top - k to 1 gives us greedy decoding. Default is 1"
        },
    )
    lm_gen_top_p: float = field(
        default=0.8,
        metadata={
            "help": "Top-p is usually set to a high value (like 0.75) with the purpose of limiting the long tail of low-probability tokens that may be sampled. We can use both top-k and top-p together. If both k and p are enabled, p acts after k. Default is 0.8."
        },
    )
    # https://github.com/huggingface/transformers/issues/27670
    lm_gen_min_p: float = field(
        default=0.0,
        metadata={
            "help": "samples from tokens with probability larger than min_p * highest_token_probability. Default is 0.0."
        },
    )
    lm_gen_repetition_penalty: float = field(
        default=1.0,
        metadata={
            "help": "Controls the token repetition pealty.  no repetition Default is 1.0, >1.0: low repetition, <1.0: high repetition"
        },
    )
    lm_gen_stops: list[str] = field(
        default_factory=list,
        metadata={
            "help": "A list of strings that will stop the generation. Default is []. If the stop word is a substring of the generated text, the generation will stop."
        },
    )
    lm_gen_stop_ids: list[int] = field(
        default_factory=list,
        metadata={
            "help": "A list of token ids that will stop the generation. Default is []. If the stop id is a substring token id of the generated text, the generation will stop."
        },
    )
    lm_gen_end_id: int = field(
        default=0,
        metadata={
            "help": "The end token id. Default is 0. If the end id is a substring token id of the generated text, the generation will stop."
        },
    )
    lm_gen_pad_id: int = field(
        default=0,
        metadata={
            "help": "The pad token id. Default is 0. If the pad id is a substring token id of the generated text, the generation will stop."
        },
    )
    lm_gen_max_tokens_per_step: int = field(
        default=3,
        metadata={"help": "The maximum number of tokens to generate per step. Default is 3."},
    )

    def update(self, **kwargs):
        unused_kwargs = dict()
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                unused_kwargs[key] = value
        return unused_kwargs



import asyncio
import math
import os
from time import perf_counter
import modal


app = modal.App("tts-grpc")

achatbot_version = os.getenv("ACHATBOT_VERSION", "0.0.9.post4")

tts_grpc_image = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.5.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install(
        "git",
        "git-lfs",
        "ffmpeg",
        "openmpi-bin",
        "libopenmpi-dev",
        "cmake",
    )  # OpenMPI for distributed communication
    .pip_install(
        f"achatbot[{os.getenv('TTS_TAG', 'tts_generator_spark')}]=={achatbot_version}",
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    )
    .env(
        {
            "TTS_MODEL_DIR": "/root/.achatbot/models/SparkAudio/Spark-TTS-0.5B",
            "TTS_LM_TOKENIZER_DIR": "/root/.achatbot/models/SparkAudio/Spark-TTS-0.5B/LLM",
            "LLM_MODEL_NAME_OR_PATH": "/root/.achatbot/models/SparkAudio/Spark-TTS-0.5B/LLM",
            "LLM_TORCH_DTYPE": "bfloat16",
        }
    )
    .pip_install(
        f"achatbot[librosa,soundfile]=={achatbot_version}",
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    )
    # .pip_install("numpy==1.26.4", "transformers==4.48.3")
)

generator = os.getenv("GENERATOR_ENGINE", "transformers")
if generator == "transformers":
    if os.getenv("LLM_ATTN_IMPL") == "flash_attention_2":
        tts_grpc_image = tts_grpc_image.run_commands(
            "pip install flash-attn --no-build-isolation",
        )

    tts_grpc_image = tts_grpc_image.pip_install(
        f"achatbot[transformers]=={achatbot_version}",
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    ).env(
        {
            "LLM_DEVICE": "cuda" if os.getenv("IMAGE_GPU", None) else "cpu",
            "TTS_LM_GENERATOR_TAG": "llm_transformers_generator",
            "LLM_ATTN_IMPL": os.getenv("LLM_ATTN_IMPL", "eager"),  # flash_attention_2,sdpa,eager
            "LLM_DEVICE_MAP": os.getenv("LLM_DEVICE_MAP", ""),  # auto
        }
    )

if generator == "llamacpp":
    tts_grpc_image = tts_grpc_image.apt_install("clang")
    if os.getenv("IMAGE_GPU", None):
        tts_grpc_image = tts_grpc_image.run_commands(
            "find /usr/ -name 'libcuda.so.*'",
            "echo $LD_LIBRARY_PATH",
            f"LD_LIBRARY_PATH=/usr/local/cuda-12.5/compat:$LD_LIBRARY_PATH CMAKE_ARGS='-DGGML_CUDA=on' pip install llama-cpp-python --extra-index-url='https://abetlen.github.io/llama-cpp-python/whl/cu125'",
            "pip install flash-attn --no-build-isolation",
        )
    else:
        tts_grpc_image = tts_grpc_image.pip_install(
            f"achatbot[llama_cpp]=={achatbot_version}",
        )
    tts_grpc_image = tts_grpc_image.env(
        {
            # https://huggingface.co/mradermacher/SparkTTS-LLM-GGUF
            "LLM_MODEL_PATH": f"/root/.achatbot/models/mradermacher/SparkTTS-LLM-GGUF/SparkTTS-LLM.{os.getenv('QUANT','Q8_0')}.gguf",
            "TTS_LM_GENERATOR_TAG": "llm_llamacpp_generator",
            "N_GPU_LAYERS": "-1" if os.getenv("IMAGE_GPU", None) else "0",
            "FLASH_ATTN": "1" if os.getenv("IMAGE_GPU", None) else "",
            "QUANT": os.getenv("QUANT", "Q8_0"),
        }
    )


if generator == "vllm":
    tts_grpc_image = tts_grpc_image.pip_install(
        f"achatbot[vllm]=={achatbot_version}",
        "flashinfer-python==0.2.0.post2",
        "accelerate",
        extra_options="--find-links https://flashinfer.ai/whl/cu125/torch2.5/flashinfer/",
        extra_index_url="https://flashinfer.ai/whl/cu125/torch2.5/",
    ).env(
        {
            "VLLM_USE_V1": os.getenv("VLLM_USE_V1", "1"),
            "VLLM_WORKER_MULTIPROC_METHOD": "spawn",
            "TTS_LM_GENERATOR_TAG": "llm_vllm_generator",
            "LLM_TORCH_DTYPE": os.getenv("DTYPE", "bfloat16"),
            "LLM_GPU_MEMORY_UTILIZATION": os.getenv("LLM_GPU_MEMORY_UTILIZATION", "0.9"),
        }
    )

if generator == "sglang":
    tts_grpc_image = tts_grpc_image.pip_install(
        f"achatbot[sglang]=={achatbot_version}",
        extra_options="--find-links https://flashinfer.ai/whl/cu125/torch2.5/flashinfer/",
        extra_index_url="https://flashinfer.ai/whl/cu125/torch2.5/",
    ).env(
        {
            "TTS_LM_GENERATOR_TAG": "llm_sglang_generator",
            "LLM_TORCH_DTYPE": os.getenv("DTYPE", "bfloat16"),
            "LLM_GPU_MEMORY_UTILIZATION": os.getenv("LLM_GPU_MEMORY_UTILIZATION", "0.6"),
        }
    )

if generator == "trtllm" or generator == "trtllm_runner":
    tts_grpc_image = tts_grpc_image.pip_install(
        f"achatbot[trtllm]=={achatbot_version}",
        "flashinfer-python==0.2.0.post2",
        pre=True,
        extra_index_url="https://pypi.nvidia.com",
    )
    if generator == "trtllm":
        tts_grpc_image = tts_grpc_image.env(
            {
                # "TLLM_LLMAPI_BUILD_CACHE": "1",
                "TTS_LM_GENERATOR_TAG": "llm_trtllm_generator",
                "LLM_TORCH_DTYPE": os.getenv("DTYPE", "bfloat16"),
                "LLM_GPU_MEMORY_UTILIZATION": os.getenv("LLM_GPU_MEMORY_UTILIZATION", "0.7"),
            }
        )
    if generator == "trtllm_runner":
        tts_grpc_image = tts_grpc_image.env(
            {
                "TTS_LM_GENERATOR_TAG": "llm_trtllm_runner_generator",
                "LLM_TORCH_DTYPE": os.getenv("DTYPE", "bfloat16"),
                "LLM_MODEL_NAME_OR_PATH": "/root/.achatbot/trt_models/tts-spark/trt_engines_bfloat16",
            }
        )

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)

VLLM_CACHE_DIR = "/root/.cache/vllm"
vllm_cache_vol = modal.Volume.from_name("vllm-cache", create_if_missing=True)

TRT_MODEL_DIR = "/root/.achatbot/trt_models"
trt_model_vol = modal.Volume.from_name("triton_trtllm_models", create_if_missing=True)

TRT_MODEL_CACHE_DIR = "/tmp/.cache/tensorrt_llm/llmapi/"
trt_model_cache_vol = modal.Volume.from_name("triton_trtllm_cache_models", create_if_missing=True)


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    image=tts_grpc_image.pip_install("psutil").env(
        {
            "ACHATBOT_PKG": "1",
            "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
            "TORCH_CUDA_ARCH_LIST": "7.5 8.0 8.6 8.7 8.9 9.0",
            "TTS_TAG": os.getenv("TTS_TAG", "tts_generator_spark"),
            "WARMUP_TEXT": os.getenv(
                "WARMUP_TEXT",
                "hello,你好，我是机器人。|万物之始,大道至简,衍化至繁。",
            ),
            "TTS_TEXT": os.getenv(
                "TTS_TEXT",
                # "hello,你好，我是机器人。|万物之始,大道至简,衍化至繁。|君不见黄河之水天上来，奔流到海不复回。君不见高堂明镜悲白发，朝如青丝暮成雪。人生得意须尽欢，莫使金樽空对月。天生我材必有用，千金散尽还复来。",
                "hello,你好，我是机器人。|万物之始,大道至简,衍化至繁。|君不见黄河之水天上来，奔流到海不复回。君不见高堂明镜悲白发，朝如青丝暮成雪。人生得意须尽欢，莫使金樽空对月。天生我材必有用，千金散尽还复来。|PyTorch 将值组织成Tensor ， Tensor是具有丰富数据操作操作的通用 n 维数组。|Module 定义从输入值到输出值的转换，其在正向传递期间的行为由其forward成员函数指定。Module 可以包含Tensor作为参数。|例如，线性模块包含权重参数和偏差参数，其正向函数通过将输入与权重相乘并添加偏差来生成输出。|应用程序通过在自定义正向函数中将本机Module （*例如*线性、卷积等）和Function （例如relu、pool 等）拼接在一起来组成自己的Module 。|典型的训练迭代包含使用输入和标签生成损失的前向传递、用于计算参数梯度的后向传递以及使用梯度更新参数的优化器步骤。|更具体地说，在正向传递期间，PyTorch 会构建一个自动求导图来记录执行的操作。|然后，在反向传播中，它使用自动梯度图进行反向传播以生成梯度。最后，优化器应用梯度来更新参数。训练过程重复这三个步骤，直到模型收敛。",
            ),
            "CONCURRENCY_CN": os.getenv("CONCURRENCY_CN", "1"),
        }
    ),
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
        VLLM_CACHE_DIR: vllm_cache_vol,
        TRT_MODEL_DIR: trt_model_vol,
        TRT_MODEL_CACHE_DIR: trt_model_cache_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
async def run_generator():
    from achatbot.modules.speech.tts import TTSEnvInit
    from achatbot.common.interface import ITts
    from achatbot.common.types import SessionCtx
    from achatbot.common.session import Session
    from achatbot.common.logger import Logger
    from achatbot.common.utils.helper import get_device

    import numpy as np
    import librosa
    import soundfile
    import torch
    import platform

    import uuid
    import subprocess

    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    device = get_device()
    cpu = get_cup_info()
    gpu_arch = ""
    gpu_prop = None
    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties(device)
        print(gpu_prop)
        gpu_arch = f"{gpu_prop.major}.{gpu_prop.minor}"
        print(torch.cuda.memory_summary(device=device, abbreviated=True))
    else:
        print("CUDA is not available.")

    Logger.init(
        os.getenv("LOG_LEVEL", "info").upper(),
        app_name=os.getenv("TTS_TAG"),
        is_file=False,
        is_console=True,
    )

    generator_tag = os.getenv("TTS_LM_GENERATOR_TAG", "")
    concurrency_cn = int(os.getenv("CONCURRENCY_CN", "1"))
    quant = os.getenv("QUANT", "")
    tts_engine: ITts = TTSEnvInit.initTTSEngine()

    os.makedirs(ASSETS_DIR, exist_ok=True)
    result = (
        f"tts:{tts_engine.TAG}\n"
        + f"llm generator:{generator_tag}\n"
        + f"quant:{quant}\n"
        + f"cpu:{cpu}\ngpu:{gpu_prop}\n"
        + f"concurrency_cn:{concurrency_cn}\n"
    )
    file_name = f"test_{tts_engine.TAG}_{generator_tag}_{quant}_{device}_{gpu_arch}"
    stream_info = tts_engine.get_stream_info()
    print(f"stream_info:{stream_info}")

    async def generate(text, session=None, gen_global_token_ids=None):
        if session is None:
            session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
            session.ctx.state["tts_text"] = text
            session.ctx.state["temperature"] = 0.95
        res = bytearray()
        times = []
        if text == "":
            return {"text": text, "waveform": res, "times": times, "gen_global_token_ids": []}
        session.ctx.state["gen_global_token_ids"] = gen_global_token_ids
        iter = tts_engine.synthesize(session)
        start_time = perf_counter()
        i = 0
        async for chunk in iter:
            times.append(perf_counter() - start_time)
            res.extend(chunk)
            print(i, len(chunk))
            i += 1
            start_time = perf_counter()
        if len(res) == 0:
            print("no waveform data")
            return {"text": text, "waveform": res, "times": times, "gen_global_token_ids": []}
        print(
            f"generate first chunk time: {times[0]} s, {len(res)} waveform cost time: {sum(times)} s"
        )
        if "gen_global_token_ids" in session.ctx.state:
            return {
                "text": text,
                "waveform": res,
                "times": times,
                "gen_global_token_ids": session.ctx.state["gen_global_token_ids"],
            }
        return {"text": text, "waveform": res, "times": times, "gen_global_token_ids": []}

    session = Session(**SessionCtx(str(uuid.uuid4().hex)).__dict__)
    warmup_texts = os.getenv("WARMUP_TEXT").split("|")
    gen_global_token_ids = None
    for text in warmup_texts:
        session.ctx.state["tts_text"] = text
        session.ctx.state["temperature"] = 0.95
        warmup_res = await generate(text, session, gen_global_token_ids=gen_global_token_ids)
        gen_global_token_ids = warmup_res.get("gen_global_token_ids")

    texts = os.getenv("TTS_TEXT").split("|")
    iter_cn = math.ceil(len(texts) / concurrency_cn)
    for idx in range(iter_cn):
        texts_chunk = texts[idx * concurrency_cn : (idx + 1) * concurrency_cn]
        tasks = [generate(text, gen_global_token_ids=gen_global_token_ids) for text in texts_chunk]
        start_time = perf_counter()
        gather_res = await asyncio.gather(*tasks)
        tts_cost_time = perf_counter() - start_time
        assert len(gather_res) == len(texts_chunk)

        waveform_bytes = bytearray()
        for res in gather_res:
            waveform_bytes.extend(res["waveform"])
        if len(waveform_bytes) == 0:
            print("no waveform data")
            continue
        file_path = os.path.join(ASSETS_DIR, f"{file_name}_{idx}.wav")
        data = np.frombuffer(waveform_bytes, dtype=stream_info["np_dtype"])

        # 去除静音部分
        # data, _ = librosa.effects.trim(data, top_db=60)
        # todo: need eval (WER|SIM)

        soundfile.write(file_path, data, stream_info["rate"])
        info = soundfile.info(file_path, verbose=True)
        print(info)

        text = "".join(texts_chunk)
        res = f"\n{text}\n"
        res += f"tts cost time {tts_cost_time} s, wav duration {info.duration} s, RTF: {tts_cost_time/info.duration}\n"
        res += "\n" + "--" * 10 + "\n"
        print(res)
        result += res

        if torch.cuda.is_available():
            total_memory = gpu_prop.total_memory
            allocated_memory = torch.cuda.memory_allocated(device)
            memory_usage_percent = (allocated_memory / total_memory) * 100
            print(
                f"PyTorch(exclude llm generator) GPU Memory Usage: {allocated_memory / (1024 ** 2):.2f} MB / {total_memory / (1024 ** 2):.2f} MB ({memory_usage_percent:.2f}%)"
            )

            print(torch.cuda.memory_summary(device=device, abbreviated=True))
            subprocess.run("nvidia-smi", shell=True)

    file_path = os.path.join(ASSETS_DIR, f"{file_name}.log")
    with open(file_path, "w") as f:
        f.write(result)

    return (result, file_name)


def get_cup_info():
    import psutil
    import platform

    # 获取CPU信息
    cpu_count = psutil.cpu_count(logical=False)
    cpu_count_logical = psutil.cpu_count(logical=True)
    cpu_freq = psutil.cpu_freq()
    cpu_times = psutil.cpu_times()

    cpu_info = {
        "cpu_arch": platform.machine(),
        "cpu_count_physical": cpu_count,
        "cpu_count_logical": cpu_count_logical,
        "cpu_freq_current": cpu_freq.current,
        "cpu_freq_min": cpu_freq.min,
        "cpu_freq_max": cpu_freq.max,
        "cpu_times_user": cpu_times.user,
        "cpu_times_system": cpu_times.system,
        "cpu_times_idle": cpu_times.idle,
    }
    return cpu_info


"""
# tts_generator_spark

# transformers with cpu
TTS_TAG=tts_generator_spark modal run src/tts/run_generator_tts.py
# transformers with gpu cuda eager
TTS_TAG=tts_generator_spark IMAGE_GPU=T4 LLM_ATTN_IMPL=eager modal run src/tts/run_generator_tts.py
# transformers with gpu cuda sdpa
TTS_TAG=tts_generator_spark IMAGE_GPU=T4 LLM_ATTN_IMPL=sdpa modal run src/tts/run_generator_tts.py
# transformers with gpu cuda flash_attention_2
TTS_TAG=tts_generator_spark IMAGE_GPU=L4 LLM_ATTN_IMPL=flash_attention_2 modal run src/tts/run_generator_tts.py
# transformers with auto device cuda flash_attention_2
TTS_TAG=tts_generator_spark IMAGE_GPU=L4 LLM_ATTN_IMPL=flash_attention_2 LLM_DEVICE_MAP=auto modal run src/tts/run_generator_tts.py

# f16 don't support
# llamacpp with cpu, quant Q8_0
GENERATOR_ENGINE=llamacpp TTS_TAG=tts_generator_spark QUANT=Q8_0 ACHATBOT_VERSION=0.0.9.post6 modal run src/tts/run_generator_tts.py
# llamacpp with cpu, quant Q4_K_M
GENERATOR_ENGINE=llamacpp TTS_TAG=tts_generator_spark QUANT=Q4_K_M ACHATBOT_VERSION=0.0.9.post6 modal run src/tts/run_generator_tts.py
# llamacpp with cpu, quant Q2_K
GENERATOR_ENGINE=llamacpp TTS_TAG=tts_generator_spark QUANT=Q2_K ACHATBOT_VERSION=0.0.9.post6 modal run src/tts/run_generator_tts.py
# llamacpp with gpu cuda, quant Q8_0 flash attention
GENERATOR_ENGINE=llamacpp TTS_TAG=tts_generator_spark QUANT=Q8_0 ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L4 modal run src/tts/run_generator_tts.py
# llamacpp with gpu cuda, quant Q4_K_M flash attention
GENERATOR_ENGINE=llamacpp TTS_TAG=tts_generator_spark QUANT=Q4_K_M ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L4 modal run src/tts/run_generator_tts.py
# llamacpp with gpu cuda, quant Q2_K
GENERATOR_ENGINE=llamacpp TTS_TAG=tts_generator_spark QUANT=Q2_K ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=T4 modal run src/tts/run_generator_tts.py

# vllm with gpu cuda | bf16 | Using Flash Attention backend | Using FlashInfer for top-p & top-k sampling
GENERATOR_ENGINE=vllm TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L4 modal run src/tts/run_generator_tts.py
GENERATOR_ENGINE=vllm TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py

# sglang with gpu cuda | bf16 | Using FlashInfer Attention backend (flashinfer.jit)
GENERATOR_ENGINE=sglang TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py

# tensorrt-llm with gpu cuda | bf16
GENERATOR_ENGINE=trtllm TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py
# tensorrt-llm runner with gpu cuda | bf16
GENERATOR_ENGINE=trtllm_runner TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py

Tips: run trtllm_runner generator engine, if use diff gpu arch, need rebuild engine 

# CONCURRENCY_CN=4
CONCURRENCY_CN=4 GENERATOR_ENGINE=vllm TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py
CONCURRENCY_CN=4 GENERATOR_ENGINE=sglang TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py
CONCURRENCY_CN=4 GENERATOR_ENGINE=trtllm TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py
CONCURRENCY_CN=4 GENERATOR_ENGINE=trtllm_runner TTS_TAG=tts_generator_spark ACHATBOT_VERSION=0.0.9.post6 IMAGE_GPU=L40S modal run src/tts/run_generator_tts.py
"""


@app.local_entrypoint()
def main():
    result, file_name = run_generator.remote()
    with open(f"{file_name}.log", "w") as f:
        f.write(result)
